from __future__ import annotations

import asyncio
from asyncio import timeout as asyncio_timeout
from collections.abc import Callable, Coroutine
import contextlib
from dataclasses import dataclass
from datetime import UTC, datetime
import enum
import itertools
import logging
import math
import time
import typing
from typing import Any, TypeVar
import warnings

from zigpy import zdo
from zigpy.const import (
    APS_REPLY_TIMEOUT,
    APS_REPLY_TIMEOUT_EXTENDED,
    SIG_ENDPOINTS,
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    SIG_MANUFACTURER,
    SIG_MODEL,
    SIG_NODE_DESC,
)
import zigpy.datastructures
import zigpy.endpoint
import zigpy.exceptions
from zigpy.exceptions import DeliveryError
import zigpy.listeners
from zigpy.ota.manager import update_firmware
from zigpy.profiles import zha, zll
import zigpy.types as t
import zigpy.util
from zigpy.zcl import Cluster, ClusterType, OtaQueryCacheClearedEvent, foundation
from zigpy.zcl.clusters.general import Ota, PollControl, QueryNextImageCommand
import zigpy.zdo.types as zdo_t

if typing.TYPE_CHECKING:
    _R = TypeVar("_R")

    from zigpy.application import ControllerApplication
    from zigpy.ota.providers import OtaImageWithMetadata


LOGGER = logging.getLogger(__name__)

PACKET_DEBOUNCE_WINDOW = 10
MAX_DEVICE_CONCURRENCY = 2
DEFAULT_FAST_POLL_TIMEOUT = 30
DEFAULT_REQUEST_RETRIES = 2
DEFAULT_REQUEST_RETRY_DELAY = 0.1

AFTER_OTA_ATTR_READ_DELAY = 10


@dataclass(frozen=True, slots=True)
class ResponseKey:
    """Key for request/response matching."""

    endpoint_id: int
    cluster_id: int
    direction: foundation.Direction | None
    tsn: int


class Status(enum.IntEnum):
    """The status of a Device. Maintained for backwards compatibility."""

    # No initialization done
    NEW = 0
    # ZDO endpoint discovery done
    ZDO_INIT = 1
    # Endpoints initialized
    ENDPOINTS_INIT = 2


class Device(zigpy.util.LocalLogMixin, zigpy.util.ListenableMixin):
    """A device on the network"""

    manufacturer_id_override = None

    def __init__(self, application: ControllerApplication, ieee: t.EUI64, nwk: t.NWK):
        self._application: ControllerApplication = application
        self._ieee: t.EUI64 = ieee
        self.nwk: t.NWK = t.NWK(nwk)
        self.zdo: zdo.ZDO = zdo.ZDO(self)
        self.endpoints: dict[int, zdo.ZDO | zigpy.endpoint.Endpoint] = {0: self.zdo}

        # Persist the original signature for the device, before quirks are applied
        self._original_signature: dict[str, Any] | None = None

        self.lqi: int | None = None
        self.rssi: int | None = None
        self.ota_in_progress: bool = False
        self._last_seen: datetime | None = None

        self._initialize_task: asyncio.Task | None = None
        self._reinterview_in_progress: bool = False
        self._group_scan_task: asyncio.Task | None = None
        self._fast_polling_reset_task: asyncio.Task | None = None

        self._listeners = {}
        self._manufacturer: str | None = None
        self._model: str | None = None
        self.node_desc: zdo_t.NodeDescriptor | None = None
        self._requests: dict[ResponseKey, asyncio.Future] = {}
        self._relays: t.Relays | None = None
        self._skip_configuration: bool = False
        self._send_sequence: int = 0

        self._fast_polling = False
        self._on_remove_callbacks: list[typing.Callable[[], None]] = []
        self._tasks: set[asyncio.Future[Any]] = set()

        self._packet_debouncer = zigpy.datastructures.Debouncer()
        self._concurrent_requests_semaphore = zigpy.datastructures.RequestLimiter(
            max_concurrency=MAX_DEVICE_CONCURRENCY,
            capacities={
                t.PacketPriority.HIGH: 0.5,
                # t.PacketPriority.NORMAL is shared with LOW
                t.PacketPriority.LOW: 0.5,
            },
        )

        # Retained for backwards compatibility, will be removed in a future release
        self.status = Status.NEW

        self._on_remove_callbacks.append(
            self._application.register_callback_listener(
                src=self,
                filters=[PollControl.ClientCommandDefs.checkin.schema()],
                callback=self.poll_control_checkin_callback,
            )
        )

    def create_task(
        self, target: Coroutine[Any, Any, _R], name: str | None = None
    ) -> asyncio.Task[_R]:
        """Create a task and store a reference to it until the task completes.

        target: target to call.
        """
        task = asyncio.get_running_loop().create_task(target, name=name)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.remove)
        return task

    def on_remove(self) -> None:
        """Call on remove callbacks."""
        for callback in self._on_remove_callbacks:
            callback()

        self._on_remove_callbacks.clear()

        for task in self._tasks:
            task.cancel()

        self._tasks.clear()

    @contextlib.asynccontextmanager
    async def _limit_concurrency(self, *, priority: int | None = None):
        """Async context manager to limit device request concurrency."""
        # Defer to the current app-level priority if not specified
        if priority is None:
            priority = self._application._packet_priority_var.get()

        start_time = time.monotonic()
        manager: contextlib.AbstractAsyncContextManager

        if priority >= t.PacketPriority.CRITICAL:
            LOGGER.debug(
                "Critical priority request received (%s), skipping queue with %d requests",
                priority,
                self._concurrent_requests_semaphore.waiting_requests,
            )
            manager = contextlib.nullcontext()
            was_locked = False
        else:
            manager = self._concurrent_requests_semaphore(priority=priority)
            was_locked = self._concurrent_requests_semaphore.locked(priority=priority)

        if was_locked:
            LOGGER.debug(
                "Device concurrency (%s) reached, delaying request (%s enqueued)",
                self._concurrent_requests_semaphore.active_requests,
                self._concurrent_requests_semaphore.waiting_requests,
            )

        async with manager:
            if was_locked:
                LOGGER.debug(
                    "Previously delayed device request is now running, delayed by %0.2fs",
                    time.monotonic() - start_time,
                )

            yield

    def get_sequence(self) -> int:
        self._send_sequence = (self._send_sequence + 1) % 256
        return self._send_sequence

    @property
    def name(self) -> str:
        return f"0x{self.nwk:04X}"

    def update_last_seen(self) -> None:
        """Update the `last_seen` attribute to the current time and emit an event."""

        warnings.warn(
            "Calling `update_last_seen` directly is deprecated", DeprecationWarning
        )

        self.last_seen = datetime.now(UTC)

    @property
    def last_seen(self) -> float | None:
        return self._last_seen.timestamp() if self._last_seen is not None else None

    @last_seen.setter
    def last_seen(self, value: datetime | float | None):
        if isinstance(value, int | float):
            value = datetime.fromtimestamp(value, UTC)

        self._last_seen = value
        self.listener_event("device_last_seen_updated", self._last_seen)

    @property
    def non_zdo_endpoints(self) -> list[zigpy.endpoint.Endpoint]:
        return [
            ep for epid, ep in self.endpoints.items() if not (isinstance(ep, zdo.ZDO))
        ]

    @property
    def has_non_zdo_endpoints(self) -> bool:
        return bool(self.non_zdo_endpoints)

    @property
    def all_endpoints_init(self) -> bool:
        return self.has_non_zdo_endpoints and all(
            ep.status != zigpy.endpoint.Status.NEW for ep in self.non_zdo_endpoints
        )

    @property
    def is_initialized(self) -> bool:
        return self.node_desc is not None and self.all_endpoints_init

    def schedule_group_membership_scan(self) -> asyncio.Task:
        """Rescan device group's membership."""
        if self._group_scan_task and not self._group_scan_task.done():
            self.debug("Cancelling old group rescan")
            self._group_scan_task.cancel()

        self._group_scan_task = self.create_task(
            self.group_membership_scan(), name="group_membership_scan"
        )
        return self._group_scan_task

    async def group_membership_scan(self) -> None:
        """Sync up group membership."""
        for ep in self.non_zdo_endpoints:
            await ep.group_membership_scan()

    @property
    def initializing(self) -> bool:
        """Return True if device is being initialized."""
        return self._initialize_task is not None and not self._initialize_task.done()

    @property
    def reinterviewing(self) -> bool:
        """Return True if device is being re-interviewed."""
        return self._reinterview_in_progress

    def cancel_initialization(self) -> None:
        """Cancel initialization call."""
        if self.initializing:
            self.debug("Canceling old initialize call")
            self._initialize_task.cancel()  # type:ignore[union-attr]

    def schedule_initialize(self) -> asyncio.Task | None:
        # Already-initialized devices don't need to be re-initialized
        if self.is_initialized:
            self.debug("Skipping initialization, device is fully initialized")
            self._application.device_initialized(self)
            return None

        if self.reinterviewing:
            self.debug("Skipping initialization, re-interview in progress")
            return None

        self.debug("Scheduling initialization")

        self.cancel_initialization()
        self._initialize_task = self.create_task(self.initialize(), name="initialize")

        return self._initialize_task

    async def reinterview(self) -> None:
        """Re-interview this device using a shadow device for safety.

        Creates a fresh Device, discovers its endpoints/clusters/model info,
        and on success swaps the old device out. On failure, the old device
        and its DB state are preserved.
        """
        if self.reinterviewing:
            self.debug("Re-interview already in progress, skipping")
            return

        if self.initializing:
            self.debug("Initialization in progress, skipping re-interview")
            return

        if self.ota_in_progress:
            self.debug("OTA in progress, skipping re-interview")
            return

        self._reinterview_in_progress = True

        try:
            shadow = Device(self._application, self._ieee, self.nwk)
            shadow._reinterview_in_progress = True  # prevent auto-initialization

            # Temporarily register the shadow in app.devices so it receives
            # ZDO responses routed by packet_received().
            self._application.devices[self._ieee] = shadow

            try:
                async with self._application.request_priority(
                    t.PacketPriority.CRITICAL
                ):
                    await shadow._discover()
            except Exception:
                # Discovery failed — restore old device, clean up shadow
                self._application.devices[self._ieee] = self
                shadow.on_remove()
                raise

            # Discovery succeeded — swap the old device for the new one.
            # If this somehow fails, the state may be partially swapped; attempting
            # to undo would likely make things worse.
            await self._application._device_reinterviewed(self, shadow)
        except Exception:  # noqa: BLE001
            self.warning(
                "Re-interview failed, keeping existing device",
                exc_info=True,
            )
            self._application.listener_event("device_reinterview_failure", self)
        finally:
            self._reinterview_in_progress = False

    async def get_node_descriptor(self) -> zdo_t.NodeDescriptor:
        self.info("Requesting 'Node Descriptor'")

        status, _, node_desc = await self.zdo.Node_Desc_req(self.nwk)

        if status != zdo_t.Status.SUCCESS:
            raise zigpy.exceptions.InvalidResponse(
                f"Requesting Node Descriptor failed: {status}"
            )

        self.node_desc = node_desc
        self.info("Got Node Descriptor: %s", node_desc)

        return node_desc

    async def initialize(self) -> None:
        try:
            # Perform initialization with critical priority
            async with self._application.request_priority(t.PacketPriority.CRITICAL):
                await self._initialize()
        except (TimeoutError, zigpy.exceptions.ZigbeeException):
            self.application.listener_event("device_init_failure", self)
        except Exception:  # noqa: BLE001
            LOGGER.warning(
                "Device %r failed to initialize due to unexpected error",
                self,
                exc_info=True,
            )

            self.application.listener_event("device_init_failure", self)

    def find_cluster(
        self, cluster_id: int, cluster_type: ClusterType = ClusterType.Server
    ) -> Cluster:
        """Find the first cluster by its ID and type on any endpoint."""
        for ep in self.non_zdo_endpoints:
            if cluster_type == ClusterType.Server and cluster_id in ep.in_clusters:
                return ep.in_clusters[cluster_id]
            elif cluster_type == ClusterType.Client and cluster_id in ep.out_clusters:
                return ep.out_clusters[cluster_id]
        raise ValueError(
            f"Cluster {cluster_id:#06x} not found in any endpoint of device {self}"
        )

    async def poll_control_checkin_callback(
        self,
        zcl_hdr: foundation.ZCLHeader,
        command: foundation.CommandSchema,
    ) -> None:
        """Handle Poll Control check-in callback."""
        poll_control = self.find_cluster(cluster_id=PollControl.cluster_id)

        async with self._application.request_priority(t.PacketPriority.CRITICAL):
            # Initiate fast polling mode if we are initializing or waiting for requests
            # to be sent
            if (
                self.initializing
                or self.reinterviewing
                or self._concurrent_requests_semaphore.active_requests > 0
                or self._fast_polling
            ):
                # Initiate fast polling mode if we are initializing or waiting for
                # requests to be sent
                await poll_control.checkin_response(
                    start_fast_polling=True,
                    fast_poll_timeout=int(DEFAULT_FAST_POLL_TIMEOUT * 4),
                    tsn=zcl_hdr.tsn,
                    expect_reply=False,
                    disable_default_response=True,
                )
            else:
                await poll_control.checkin_response(
                    start_fast_polling=False,
                    fast_poll_timeout=0,
                    tsn=zcl_hdr.tsn,
                    expect_reply=False,
                    disable_default_response=True,
                )

    async def begin_fast_polling(
        self, timeout: float = DEFAULT_FAST_POLL_TIMEOUT, *, reset_after: bool = True
    ) -> None:
        """Ask the device to enter fast polling mode."""
        try:
            poll_control = self.find_cluster(cluster_id=PollControl.cluster_id)
        except ValueError:
            LOGGER.debug("Device does not support fast polling")
            # The device doesn't have the cluster, there's nothing more we can do
            return

        # Cancel any fast poll reset tasks
        if self._fast_polling_reset_task is not None:
            self._fast_polling_reset_task.cancel()
            self._fast_polling_reset_task = None

        # The units are quarter seconds, we round up to the nearest one
        adjusted_timeout = math.ceil(timeout * 4) / 4
        LOGGER.debug("Beginning fast polling for %0.2fs", adjusted_timeout)

        # We must first bind to the cluster, otherwise the device will not send a check-
        # in command
        await poll_control.bind()
        await poll_control.write_attributes(
            {PollControl.AttributeDefs.fast_poll_timeout.id: int(4 * adjusted_timeout)}
        )

        self._fast_polling = True

        if reset_after:

            async def reset_fast_polling() -> None:
                await asyncio.sleep(adjusted_timeout)
                self._fast_polling = False

            self._fast_polling_reset_task = self.create_task(
                reset_fast_polling(), name="reset_fast_polling"
            )

    @contextlib.asynccontextmanager
    async def fast_poll_mode(
        self, initial_timeout: float = DEFAULT_FAST_POLL_TIMEOUT
    ) -> None:
        """Ask the device to enter fast polling mode."""
        await self.begin_fast_polling(timeout=initial_timeout, reset_after=False)

        try:
            yield
        finally:
            LOGGER.debug("Stopping fast polling on next device check-in")
            self._fast_polling = False

    async def _discover(self) -> None:
        """Discover all basic information about a device: its node descriptor, all
        endpoints and clusters, and the model and manufacturer attributes from any
        Basic cluster exposing those attributes. Does not signal completion.
        """

        # Some devices are improperly initialized and are missing a node descriptor
        if self.node_desc is None:
            await self.get_node_descriptor()

        # Devices should have endpoints other than ZDO
        if self.has_non_zdo_endpoints:
            self.info("Already have endpoints: %s", self.endpoints)
        else:
            self.info("Discovering endpoints")

            status, _, endpoints = await self.zdo.Active_EP_req(self.nwk)

            if status != zdo_t.Status.SUCCESS:
                raise zigpy.exceptions.InvalidResponse(
                    f"Endpoint request failed: {status}"
                )

            self.info("Discovered endpoints: %s", endpoints)

            for endpoint_id in endpoints:
                if endpoint_id != 0:
                    self.add_endpoint(endpoint_id)

        self.status = Status.ZDO_INIT

        # Initialize all of the discovered endpoints
        initiated_fast_polling = self._fast_polling

        if self.all_endpoints_init:
            self.info(
                "All endpoints are already initialized: %s", self.non_zdo_endpoints
            )

            if not initiated_fast_polling:
                # Begin fast polling if we are re-initializing
                await self.begin_fast_polling()
        else:
            self.info("Initializing endpoints %s", self.non_zdo_endpoints)

            for ep in self.non_zdo_endpoints:
                await ep.initialize()

                if not initiated_fast_polling:
                    # Ask the device to enter fast polling mode as soon as we are
                    # aware of a PollControl cluster
                    try:
                        await self.begin_fast_polling()
                    except (TimeoutError, DeliveryError):
                        pass
                    else:
                        initiated_fast_polling = True

        # Query model info
        if self.model is not None and self.manufacturer is not None:
            self.info("Already have model and manufacturer info")
        else:
            for ep in self.non_zdo_endpoints:
                if self.model is None or self.manufacturer is None:
                    model, manufacturer = await ep.get_model_info()
                    self.info(
                        "Read model %r and manufacturer %r from %s",
                        model,
                        manufacturer,
                        ep,
                    )

                    if model is not None:
                        self.model = model

                    if manufacturer is not None:
                        self.manufacturer = manufacturer

        try:
            ota = self.find_cluster(
                cluster_id=Ota.cluster_id, cluster_type=ClusterType.Client
            )
        except ValueError:
            self.debug("Device does not support OTA cluster")
        else:
            await ota.read_attributes([Ota.AttributeDefs.current_file_version.name])

            # Notify the device that OTA images may be available so it queries us
            try:
                await ota.image_notify(
                    payload_type=Ota.ImageNotifyCommand.PayloadType.QueryJitter,
                    query_jitter=100,
                )
            except Exception:  # noqa: BLE001
                self.debug(
                    "OTA image_notify failed during initialization", exc_info=True
                )

        self.status = Status.ENDPOINTS_INIT

        self.info("Discovered basic device information for %s", self)

    async def _initialize(self) -> None:
        """Discover device information and signal to the application."""
        await self._discover()

        # Signal to the application that the device is ready
        self._application.device_initialized(self)

    def add_endpoint(self, endpoint_id) -> zigpy.endpoint.Endpoint:
        ep = zigpy.endpoint.Endpoint(self, endpoint_id)
        self.endpoints[endpoint_id] = ep
        return ep

    async def add_to_group(self, grp_id: int, name: str | None = None) -> None:
        for ep in self.non_zdo_endpoints:
            await ep.add_to_group(grp_id, name)

    async def remove_from_group(self, grp_id: int) -> None:
        for ep in self.non_zdo_endpoints:
            await ep.remove_from_group(grp_id)

    async def request(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        timeout=APS_REPLY_TIMEOUT,
        use_ieee=False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_delay: float | None = None,
        **kwargs,
    ):
        if retries is None:
            retries = DEFAULT_REQUEST_RETRIES

        if retry_delay is None:
            retry_delay = DEFAULT_REQUEST_RETRY_DELAY

        extended_timeout = False

        if self.node_desc is None or self.node_desc.is_end_device:
            self.debug("Extending timeout for 0x%02x request", sequence)
            timeout = APS_REPLY_TIMEOUT_EXTENDED
            extended_timeout = True

        if expect_reply:
            if dst_ep == zdo.ZDO_ENDPOINT:
                rsp_key = ResponseKey(
                    endpoint_id=dst_ep,
                    # e.g. Node_Desc_req = 0x0002 corresponds to Node_Desc_rsp = 0x8002
                    cluster_id=cluster ^ 0x8000,
                    direction=None,
                    tsn=sequence,
                )
            else:
                zcl_hdr, _ = foundation.ZCLHeader.deserialize(data)
                rsp_key = ResponseKey(
                    endpoint_id=dst_ep,
                    cluster_id=cluster,
                    direction=zcl_hdr.frame_control.direction.flip(),
                    tsn=sequence,
                )

            if rsp_key in self._requests:
                self.debug(
                    "Duplicate request key %s, pending requests %s",
                    rsp_key,
                    self._requests,
                )
                raise zigpy.exceptions.ControllerException(
                    f"Duplicate request key: {rsp_key}"
                )
        else:
            rsp_key = None

        max_attempts = retries + 1

        # Use a lambda so we don't leave the coroutine unawaited in case of an exception
        send_request = lambda attempt: self._application.request(  # noqa: E731
            device=self,
            profile=profile,
            cluster=cluster,
            src_ep=src_ep,
            dst_ep=dst_ep,
            sequence=sequence,
            data=data,
            expect_reply=expect_reply,
            use_ieee=use_ieee,
            extended_timeout=extended_timeout,
            ask_for_ack=ask_for_ack,
            priority=priority,
            force_route_discovery=(attempt > 0),
            **kwargs,
        )

        for attempt in range(max_attempts):
            try:
                async with self._limit_concurrency(priority=priority):
                    if not expect_reply:
                        await send_request(attempt=attempt)
                        return None

                    assert rsp_key is not None

                    future: asyncio.Future[list[Any] | foundation.CommandSchema] = (
                        asyncio.Future()
                    )
                    self._requests[rsp_key] = future

                    try:
                        await send_request(attempt=attempt)
                        async with asyncio_timeout(timeout):
                            return await future
                    finally:
                        if not future.done():
                            future.cancel()
                        self._requests.pop(rsp_key, None)
            except zigpy.exceptions.ParsingError:
                raise
            except Exception:
                LOGGER.debug(
                    "Failed to send request, attempt %d of %d",
                    attempt + 1,
                    max_attempts,
                    exc_info=True,
                )

                if attempt >= max_attempts - 1:
                    raise

                await asyncio.sleep(retry_delay)
                continue

    def handle_message(
        self,
        profile: int,
        cluster: int,
        src_ep: int,
        dst_ep: int,
        message: bytes,
    ):
        """Deprecated compatibility function. Use `packet_received` instead."""

        warnings.warn(
            "`handle_message` is deprecated, use `packet_received`", DeprecationWarning
        )

        self.packet_received(
            t.ZigbeePacket(
                profile_id=profile,
                cluster_id=cluster,
                src_ep=src_ep,
                dst_ep=dst_ep,
                data=t.SerializableBytes(message),
                dst=t.AddrModeAddress(
                    addr_mode=t.AddrMode.NWK,
                    address=self.nwk,
                ),
            )
        )

    def _find_zcl_cluster_strict(
        self, hdr: foundation.ZCLHeader, packet: t.ZigbeePacket
    ) -> Cluster:
        """Find the ZCL cluster for a given header and packet, strict."""
        assert packet.src_ep is not None
        ep = self.endpoints[packet.src_ep]

        if hdr.frame_control.direction == foundation.Direction.Client_to_Server:
            return ep.out_clusters[packet.cluster_id]
        else:
            return ep.in_clusters[packet.cluster_id]

    def _find_zcl_cluster(
        self, hdr: foundation.ZCLHeader, packet: t.ZigbeePacket
    ) -> Cluster:
        """Find the ZCL cluster for a given header and packet."""
        try:
            return self._find_zcl_cluster_strict(hdr, packet)
        except KeyError:
            # If the cluster is not found, try to find it with flipped direction. This
            # will be removed in 2025.9.0.
            cluster = self._find_zcl_cluster_strict(
                hdr.replace(
                    frame_control=hdr.frame_control.replace(
                        direction=hdr.frame_control.direction.flip()
                    )
                ),
                packet,
            )
            LOGGER.debug(
                (
                    "Cluster 0x%04x on %r has incorrect direction (got %r for %r cluster)."
                    " Please report this here: https://github.com/zigpy/zigpy/issues/1640"
                ),
                packet.cluster_id,
                self,
                hdr.frame_control.direction,
                cluster.cluster_type,
            )

            return cluster

    def custom_profile_packet_received(self, packet: t.ZigbeePacket) -> None:
        """Handle packets with a custom profile ID."""
        self.debug(
            "Received packet with custom profile 0x%04x, ignoring",
            packet.profile_id,
        )

    def _should_filter_packet(self, packet: t.ZigbeePacket) -> bool:
        """Check if packet should be filtered as duplicate."""
        return self._packet_debouncer.filter(
            # Be conservative with deduplication
            obj=packet.replace(timestamp=None, tsn=None, lqi=None, rssi=None),
            expire_in=PACKET_DEBOUNCE_WINDOW,
        )

    def _parse_packet_header(
        self, packet: t.ZigbeePacket
    ) -> tuple[zdo_t.ZDOHeader | foundation.ZCLHeader, ResponseKey] | tuple[None, None]:
        """Parse packet header and create response key."""
        data = packet.data.serialize()

        if packet.src_ep == zdo.ZDO_ENDPOINT:
            hdr, _ = zdo_t.ZDOHeader.deserialize(packet.cluster_id, data)
            rsp_key = ResponseKey(
                endpoint_id=packet.src_ep,
                cluster_id=packet.cluster_id,
                direction=None,
                tsn=hdr.tsn,
            )
            return hdr, rsp_key
        elif packet.profile_id in (zha.PROFILE_ID, zll.PROFILE_ID):
            hdr, _ = foundation.ZCLHeader.deserialize(data)
            rsp_key = ResponseKey(
                endpoint_id=packet.src_ep,
                cluster_id=packet.cluster_id,
                direction=hdr.frame_control.direction,
                tsn=hdr.tsn,
            )
            return hdr, rsp_key
        else:
            return None, None

    def _match_packet_endpoint_cluster(
        self, packet: t.ZigbeePacket, hdr: zdo_t.ZDOHeader | foundation.ZCLHeader
    ) -> (
        tuple[zigpy.endpoint.Endpoint, Cluster]
        | tuple[zigpy.zdo.ZDO, None]
        | tuple[None, None]
    ):
        """Validate packet routing and find target endpoint and cluster."""
        if packet.src_ep not in self.endpoints:
            self.debug(
                "Ignoring message on unknown endpoint %s (expected one of %s)",
                packet.src_ep,
                self.endpoints,
            )
            return None, None

        endpoint = self.endpoints[packet.src_ep]

        if packet.src_ep == zdo.ZDO_ENDPOINT:
            assert isinstance(endpoint, zigpy.zdo.ZDO)
            return endpoint, None
        else:
            assert isinstance(endpoint, zigpy.endpoint.Endpoint)
            try:
                zcl_cluster = self._find_zcl_cluster(hdr, packet)
            except KeyError:
                self.debug(
                    "Ignoring message on unknown cluster: 0x%04x",
                    packet.cluster_id,
                )
                return None, None
            else:
                return endpoint, zcl_cluster

    def _parse_packet_command(
        self, packet: t.ZigbeePacket, endpoint: typing.Any, zcl_cluster: Cluster | None
    ) -> typing.Any:
        """Deserialize packet data."""
        data = packet.data.serialize()

        if packet.src_ep == zdo.ZDO_ENDPOINT:
            _, cmd = endpoint.deserialize(packet.cluster_id, data)
        else:
            assert zcl_cluster is not None
            _, cmd = zcl_cluster.deserialize(data)

        return cmd

    def _maybe_match_response(
        self, rsp_key: ResponseKey, cmd: typing.Any | None, error: Exception | None
    ) -> bool:
        """Handle response matching for pending requests, returns True if packet was matched."""
        future = self._requests.get(rsp_key)
        if future is None:
            return False

        # Attribute reports often collide with command responses, they should never be
        # matched up
        if isinstance(
            cmd,
            foundation.GENERAL_COMMANDS[
                foundation.GeneralCommand.Report_Attributes
            ].schema,
        ):
            return False

        try:
            if error is not None:
                future.set_exception(error)
            else:
                future.set_result(cmd)
        except asyncio.InvalidStateError:
            self.debug(
                "Invalid state on future for %s -- probably duplicate response",
                rsp_key,
            )

        return True

    def packet_received(self, packet: t.ZigbeePacket) -> None:
        """Process received packet through the device's packet handling pipeline."""
        self.last_seen = packet.timestamp

        if packet.lqi is not None:
            self.lqi = packet.lqi

        if packet.rssi is not None:
            self.rssi = packet.rssi

        # Filter duplicate packets
        if self._should_filter_packet(packet):
            self.debug("Filtering duplicate packet")
            return

        # Parse packet header and create response key
        hdr, rsp_key = self._parse_packet_header(packet)
        if hdr is None:
            self.custom_profile_packet_received(packet)
            return

        # Validate packet routing and find target endpoint/cluster
        endpoint, zcl_cluster = self._match_packet_endpoint_cluster(packet, hdr)
        if endpoint is None:
            return

        # Deserialize packet data
        try:
            cmd = self._parse_packet_command(packet, endpoint, zcl_cluster)
        except Exception as exc:  # noqa: BLE001
            cmd = None
            error = zigpy.exceptions.ParsingError()
            error.__cause__ = exc
            self.debug("Failed to parse packet %r", packet, exc_info=error)
        else:
            error = None

        # Handle response matching for pending requests
        if self._maybe_match_response(rsp_key, cmd, error):
            return

        # Skip further processing if there was a parsing error
        if error is not None:
            return

        # Pass the request off to a listener, if one is registered
        for listener in itertools.chain(
            self._application._req_listeners[zigpy.listeners.ANY_DEVICE],
            self._application._req_listeners[self],
        ):
            # Resolve only until the first future listener
            if listener.resolve(hdr, cmd) and isinstance(
                listener, zigpy.listeners.FutureListener
            ):
                break

        # Finally, pass it off to the cluster message handler. This will be removed.
        if zcl_cluster is not None:
            zcl_cluster.handle_message(hdr, cmd)
        else:
            assert isinstance(endpoint, zdo.ZDO)
            endpoint.handle_message(packet.profile_id, packet.cluster_id, hdr, cmd)

    async def reply(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        timeout=APS_REPLY_TIMEOUT,
        expect_reply: bool = False,
        use_ieee: bool = False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        **kwargs,
    ):
        return await self.request(
            profile=profile,
            cluster=cluster,
            src_ep=src_ep,
            dst_ep=dst_ep,
            sequence=sequence,
            data=data,
            expect_reply=expect_reply,
            timeout=timeout,
            use_ieee=use_ieee,
            ask_for_ack=ask_for_ack,
            priority=priority,
            **kwargs,
        )

    async def update_firmware(
        self,
        image: OtaImageWithMetadata,
        progress_callback: Callable[[int, int, float], None] | None = None,
        force: bool = False,
    ) -> foundation.Status | None:
        """Update device firmware."""
        if self.ota_in_progress:
            self.debug("OTA already in progress")
            return None

        self.ota_in_progress = True

        try:
            result = await update_firmware(
                device=self,
                image=image,
                progress_callback=progress_callback,
                force=force,
            )
        except Exception as exc:  # noqa: BLE001
            self.debug("OTA failed!", exc_info=exc)
            raise
        finally:
            self.ota_in_progress = False

        if result != foundation.Status.SUCCESS:
            return result

        # Clear the current file version when the update succeeds
        ota = self.find_cluster(
            cluster_id=Ota.cluster_id, cluster_type=ClusterType.Client
        )
        ota.update_attribute(Ota.AttributeDefs.current_file_version.id, None)
        ota.last_query_cmd = None
        ota.emit(
            OtaQueryCacheClearedEvent.event_type,
            OtaQueryCacheClearedEvent(
                device_ieee=str(self.ieee),
                endpoint_id=ota.endpoint.endpoint_id,
            ),
        )

        await asyncio.sleep(AFTER_OTA_ATTR_READ_DELAY)
        await ota.read_attributes(
            [Ota.AttributeDefs.current_file_version.name],
            retries=10,
            retry_delay=AFTER_OTA_ATTR_READ_DELAY,
        )

        # Prompt device to send QueryNextImage with updated version for query cache
        try:
            await ota.image_notify(
                payload_type=Ota.ImageNotifyCommand.PayloadType.QueryJitter,
                query_jitter=100,
            )
        except Exception:  # noqa: BLE001
            self.debug("Post-OTA image_notify failed", exc_info=True)

        # Re-interview the device after successful OTA to pick up
        # any changes in clusters/endpoints/model and re-apply quirks.
        # reinterview() handles its own errors internally.
        await self.reinterview()

        return result

    def get_last_ota_query_cmd(self) -> QueryNextImageCommand | None:
        """Return the last cached QueryNextImageCommand, preferring client clusters."""
        for ep in self.non_zdo_endpoints:
            # Prefer client (out) clusters since runtime routing places
            # query_next_image there when both cluster types exist.
            for clusters in (ep.out_clusters, ep.in_clusters):
                cluster = clusters.get(Ota.cluster_id)
                if isinstance(cluster, Ota) and cluster.last_query_cmd is not None:
                    return cluster.last_query_cmd
        return None

    def radio_details(self, lqi=None, rssi=None) -> None:
        if lqi is not None:
            self.lqi = lqi
        if rssi is not None:
            self.rssi = rssi

    def log(self, lvl, msg, *args, **kwargs) -> None:
        msg = "[0x%04x] " + msg
        args = (self.nwk, *args)
        LOGGER.log(lvl, msg, *args, **kwargs)

    @property
    def application(self) -> ControllerApplication:
        return self._application

    @property
    def ieee(self) -> t.EUI64:
        return self._ieee

    @property
    def manufacturer(self) -> str | None:
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value) -> None:
        if isinstance(value, str):
            self._manufacturer = value

    @property
    def manufacturer_id(self) -> int | None:
        """Return manufacturer id."""
        if self.manufacturer_id_override:
            return self.manufacturer_id_override
        elif self.node_desc is not None:
            return self.node_desc.manufacturer_code
        else:
            return None

    @property
    def model(self) -> str | None:
        return self._model

    @model.setter
    def model(self, value) -> None:
        if isinstance(value, str):
            self._model = value

    @property
    def original_signature(self) -> dict[str, Any] | None:
        return self._original_signature

    @original_signature.setter
    def original_signature(self, value: dict[str, Any] | None) -> None:
        if self._original_signature is not None:
            return

        self._original_signature = value

    @property
    def skip_configuration(self) -> bool:
        return self._skip_configuration

    @skip_configuration.setter
    def skip_configuration(self, should_skip_configuration) -> None:
        if isinstance(should_skip_configuration, bool):
            self._skip_configuration = should_skip_configuration
        else:
            self._skip_configuration = False

    @property
    def relays(self) -> t.Relays | None:
        """Relay list."""
        return self._relays

    @relays.setter
    def relays(self, relays: t.Relays | None) -> None:
        if relays is None:
            pass
        elif not isinstance(relays, t.Relays):
            relays = t.Relays(relays)

        self._relays = relays
        self.listener_event("device_relays_updated", relays)

    def __getitem__(self, key):
        return self.endpoints[key]

    def clone(self) -> Device:
        """Return a detached copy of this device for independent modification."""
        new = Device(self.application, self.ieee, self.nwk)
        new.lqi = self.lqi
        new.rssi = self.rssi
        new.last_seen = self.last_seen
        new.relays = self.relays
        new.original_signature = self.original_signature
        new.status = self.status
        new.node_desc = self.node_desc
        new.manufacturer = self.manufacturer
        new.model = self.model
        new.manufacturer_id_override = self.manufacturer_id_override
        new.skip_configuration = self.skip_configuration

        for endpoint in self.non_zdo_endpoints:
            new_ep = new.add_endpoint(endpoint.endpoint_id)
            new_ep.status = endpoint.status
            new_ep.profile_id = endpoint.profile_id
            new_ep.device_type = endpoint.device_type

            for cluster in endpoint.in_clusters.values():
                new_cluster = new_ep.add_input_cluster(cluster.cluster_id)
                new_cluster._attr_cache_internal = cluster._attr_cache.clone(
                    new_cluster
                )

            for cluster in endpoint.out_clusters.values():
                new_cluster = new_ep.add_output_cluster(cluster.cluster_id)
                new_cluster._attr_cache_internal = cluster._attr_cache.clone(
                    new_cluster
                )

        return new

    def get_signature(self) -> dict[str, typing.Any]:
        # return the device signature by providing essential device information
        #    - Model Identifier ( Attribute 0x0005 of Basic Cluster 0x0000 )
        #    - Manufacturer Name ( Attribute 0x0004 of Basic Cluster 0x0000 )
        #    - Endpoint list
        #        - Profile Id, Device Id, Cluster Out, Cluster In
        signature: dict[str, typing.Any] = {}
        if self._manufacturer is not None:
            signature[SIG_MANUFACTURER] = self.manufacturer
        if self._model is not None:
            signature[SIG_MODEL] = self._model
        if self.node_desc is not None:
            signature[SIG_NODE_DESC] = self.node_desc.as_dict()

        for endpoint_id, endpoint in self.endpoints.items():
            if endpoint_id == zdo.ZDO_ENDPOINT:  # ZDO
                continue
            signature.setdefault(SIG_ENDPOINTS, {})
            in_clusters = list(endpoint.in_clusters)
            out_clusters = list(endpoint.out_clusters)
            signature[SIG_ENDPOINTS][endpoint_id] = {
                SIG_EP_PROFILE: endpoint.profile_id,
                SIG_EP_TYPE: endpoint.device_type,
                SIG_EP_INPUT: in_clusters,
                SIG_EP_OUTPUT: out_clusters,
            }
        return signature

    def __repr__(self) -> str:
        return (
            f"<"
            f"{type(self).__name__}"
            f" model={self.model!r}"
            f" manuf={self.manufacturer!r}"
            f" nwk={t.NWK(self.nwk)}"
            f" ieee={self.ieee}"
            f" is_initialized={self.is_initialized}"
            f">"
        )


async def broadcast(
    app,
    profile,
    cluster,
    src_ep,
    dst_ep,
    grpid,
    radius,
    sequence,
    data,
    broadcast_address=t.BroadcastAddress.RX_ON_WHEN_IDLE,
):
    return await app.broadcast(
        profile,
        cluster,
        src_ep,
        dst_ep,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=broadcast_address,
    )



TRADFRI_SCHEMA = {
    "type": "array",
    "items": {
        "oneOf": [
            {
                "type": "object",
                "properties": {
                    "fw_image_type": {"type": "integer"},
                    "fw_type": {"type": "integer"},
                    "fw_sha3_256": {"type": "string", "pattern": "^[a-f0-9]{64}$"},
                    "fw_binary_url": {"type": "string", "format": "uri"},
                },
                "required": [
                    "fw_image_type",
                    "fw_type",
                    "fw_sha3_256",
                    "fw_binary_url",
                ],
            },
            {
                "type": "object",
                "properties": {
                    # For the gateway firmware, ignore the rest of the fields
                    "fw_type": {"type": "integer", "const": 3},
                },
                "required": [
                    "fw_type",
                ],
            },
            # Old IKEA format (new gateway)
            {
                "type": "object",
                "properties": {
                    "fw_binary_url": {"type": "string", "format": "uri"},
                    "fw_filesize": {"type": "integer"},
                    "fw_hotfix_version": {"type": "integer"},
                    "fw_major_version": {"type": "integer"},
                    "fw_minor_version": {"type": "integer"},
                    "fw_req_hotfix_version": {"type": "integer"},
                    "fw_req_major_version": {"type": "integer"},
                    "fw_req_minor_version": {"type": "integer"},
                    "fw_type": {"const": 0},
                    "fw_update_prio": {"type": "integer"},
                    "fw_weblink_relnote": {"type": "string", "format": "uri"},
                },
                "required": [
                    "fw_binary_url",
                    "fw_filesize",
                    "fw_hotfix_version",
                    "fw_major_version",
                    "fw_minor_version",
                    "fw_req_hotfix_version",
                    "fw_req_major_version",
                    "fw_req_minor_version",
                    "fw_type",
                    "fw_update_prio",
                    "fw_weblink_relnote",
                ],
            },
            # Old IKEA format (device)
            {
                "type": "object",
                "properties": {
                    "fw_binary_url": {"type": "string", "format": "uri"},
                    "fw_file_version_LSB": {"type": "integer"},
                    "fw_file_version_MSB": {"type": "integer"},
                    "fw_filesize": {"type": "integer"},
                    "fw_image_type": {"type": "integer"},
                    "fw_manufacturer_id": {"type": "integer"},
                    "fw_type": {"const": 2},
                },
                "required": [
                    "fw_binary_url",
                    "fw_file_version_LSB",
                    "fw_file_version_MSB",
                    "fw_filesize",
                    "fw_image_type",
                    "fw_manufacturer_id",
                    "fw_type",
                ],
            },
            # Old IKEA format (old gateway)
            {
                "type": "object",
                "properties": {
                    "fw_binary_url": {"type": "string", "format": "uri"},
                    "fw_build_version": {"type": "integer"},
                    "fw_file_version_LSB": {"type": "integer"},
                    "fw_file_version_MSB": {"type": "integer"},
                    "fw_filesize": {"type": "integer"},
                    "fw_hotfix_version": {"type": "integer"},
                    "fw_image_type": {"type": "integer"},
                    "fw_major_version": {"type": "integer"},
                    "fw_manufacturer_id": {"type": "integer"},
                    "fw_minor_version": {"type": "integer"},
                    "fw_type": {"const": 1},
                },
                "required": [
                    "fw_binary_url",
                    "fw_build_version",
                    "fw_file_version_LSB",
                    "fw_file_version_MSB",
                    "fw_filesize",
                    "fw_hotfix_version",
                    "fw_image_type",
                    "fw_major_version",
                    "fw_manufacturer_id",
                    "fw_minor_version",
                    "fw_type",
                ],
            },
        ]
    },
}

LEDVANCE_SCHEMA = {
    "type": "object",
    "properties": {
        "firmwares": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "blob": {"type": ["null", "string"]},
                    "identity": {
                        "type": "object",
                        "properties": {
                            "company": {"type": "integer"},
                            "product": {"type": "integer"},
                            "version": {
                                "type": "object",
                                "properties": {
                                    "major": {"type": "integer"},
                                    "minor": {"type": "integer"},
                                    "build": {"type": "integer"},
                                    "revision": {"type": "integer"},
                                },
                                "required": ["major", "minor", "build", "revision"],
                            },
                        },
                        "required": ["company", "product", "version"],
                    },
                    "releaseNotes": {"type": "string"},
                    "shA256": {"type": "string", "pattern": "^[a-f0-9]{64}$"},
                    "name": {"type": "string"},
                    "productName": {"type": "string"},
                    "fullName": {"type": "string"},
                    "extension": {"type": "string"},
                    "released": {"type": "string", "format": "date-time"},
                    "salesRegion": {"type": ["string", "null"]},
                    "length": {"type": "integer"},
                },
                "required": [
                    "blob",
                    "identity",
                    "releaseNotes",
                    "shA256",
                    "name",
                    "productName",
                    "fullName",
                    "extension",
                    "released",
                    "salesRegion",
                    "length",
                ],
            },
        }
    },
    "required": ["firmwares"],
}

SONOFF_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "fw_binary_url": {"type": "string", "format": "uri"},
            "fw_file_version": {"type": "integer"},
            "fw_filesize": {"type": "integer"},
            "fw_image_type": {"type": "integer"},
            "fw_manufacturer_id": {"type": "integer"},
            "model_id": {"type": "string"},
        },
        "required": [
            "fw_binary_url",
            "fw_file_version",
            "fw_filesize",
            "fw_image_type",
            "fw_manufacturer_id",
            "model_id",
        ],
    },
}

REMOTE_PROVIDER_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "firmwares": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "binary_url": {"type": "string", "format": "uri"},
                    "path": {"type": "string"},
                    "file_version": {"type": "integer"},
                    "file_size": {"type": "integer"},
                    "image_type": {"type": "integer"},
                    "manufacturer_names": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "model_names": {"type": "array", "items": {"type": "string"}},
                    "manufacturer_id": {"type": "integer"},
                    "changelog": {"type": "string"},
                    "release_notes": {"type": "string"},
                    "release_url": {"type": "string", "format": "uri"},
                    "checksum": {
                        "type": "string",
                        "pattern": "^sha3-256:[a-f0-9]{64}$",
                    },
                    "min_hardware_version": {"type": "integer"},
                    "max_hardware_version": {"type": "integer"},
                    "min_current_file_version": {"type": "integer"},
                    "max_current_file_version": {"type": "integer"},
                    "specificity": {"type": "integer"},
                },
                "required": [
                    # "binary_url",
                    # "path",
                    "file_version",
                    "file_size",
                    "image_type",
                    # "manufacturer_names",
                    # "model_names",
                    "manufacturer_id",
                    # "changelog",
                    "checksum",
                    # "min_hardware_version",
                    # "max_hardware_version",
                    # "min_current_file_version",
                    # "max_current_file_version",
                    # "release_notes",
                    # "release_url",
                    # "specificity",
                ],
            },
        }
    },
    "required": ["firmwares"],
}

Z2M_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "fileVersion": {"type": "integer"},
            "fileSize": {"type": "integer"},
            "manufacturerCode": {"type": "integer"},
            "imageType": {"type": "integer"},
            "sha512": {"type": "string", "pattern": "^[a-f0-9]{128}$"},
            "url": {"type": "string", "format": "uri"},
            "path": {"type": "string"},
            "minFileVersion": {"type": "integer"},
            "maxFileVersion": {"type": "integer"},
            "manufacturerName": {"type": "array", "items": {"type": "string"}},
            "modelId": {"type": "string"},
        },
        "required": [
            "fileVersion",
            "fileSize",
            "manufacturerCode",
            "imageType",
            "sha512",
            "url",
        ],
    },
}



from datetime import UTC, datetime, timedelta
import json
import logging

import pytest

from tests.async_mock import AsyncMock
from tests.conftest import app  # noqa: F401
import zigpy.backups
import zigpy.state as app_state
import zigpy.types as t
import zigpy.zdo.types as zdo_t


@pytest.fixture
def backup_factory():
    def inner():
        return zigpy.backups.NetworkBackup(
            backup_time=datetime(2021, 2, 8, 19, 35, 24, 761000, tzinfo=UTC),
            node_info=app_state.NodeInfo(
                nwk=t.NWK(0x0000),
                ieee=t.EUI64.convert("93:2C:A9:34:D9:D0:5D:12"),
                logical_type=zdo_t.LogicalType.Coordinator,
                model="Coordinator Model",
                manufacturer="Coordinator Manufacturer",
                version="1.2.3.4",
            ),
            network_info=app_state.NetworkInfo(
                extended_pan_id=t.ExtendedPanId.convert("0D:49:91:99:AE:CD:3C:35"),
                pan_id=t.PanId(0x9BB0),
                nwk_update_id=0x12,
                nwk_manager_id=t.NWK(0x0000),
                channel=t.uint8_t(15),
                channel_mask=t.Channels.from_channel_list([15, 20, 25]),
                tx_power=8,
                security_level=t.uint8_t(5),
                network_key=app_state.Key(
                    key=t.KeyData.convert(
                        "9A:79:D6:9A:DA:EC:45:C6:F2:EF:EB:AF:DA:A3:07:B6"
                    ),
                    seq=108,
                    tx_counter=39009277,
                ),
                tc_link_key=app_state.Key(
                    key=t.KeyData(b"ZigBeeAlliance09"),
                    partner_ieee=t.EUI64.convert("93:2C:A9:34:D9:D0:5D:12"),
                    tx_counter=8712428,
                ),
                key_table=[
                    app_state.Key(
                        key=t.KeyData.convert(
                            "85:7C:05:00:3E:76:1A:F9:68:9A:49:41:6A:60:5C:76"
                        ),
                        tx_counter=3792973670,
                        rx_counter=1083290572,
                        seq=147,
                        partner_ieee=t.EUI64.convert("69:0C:07:52:AA:D7:7D:71"),
                    ),
                    app_state.Key(
                        key=t.KeyData.convert(
                            "CA:02:E8:BB:75:7C:94:F8:93:39:D3:9C:B3:CD:A7:BE"
                        ),
                        tx_counter=2597245184,
                        rx_counter=824424412,
                        seq=19,
                        partner_ieee=t.EUI64.convert("A3:1A:F6:8E:19:95:23:BE"),
                    ),
                ],
                children=[
                    # Has a key
                    t.EUI64.convert("A3:1A:F6:8E:19:95:23:BE"),
                    # Random device with no NWK address or key
                    t.EUI64.convert("A4:02:A0:DC:17:D8:17:DF"),
                    # Does not have a key
                    t.EUI64.convert("C6:DF:28:F9:60:33:DB:03"),
                ],
                # If exposed by the stack, NWK addresses of other connected devices on the network
                nwk_addresses={
                    # Two children above
                    t.EUI64.convert("A3:1A:F6:8E:19:95:23:BE"): t.NWK(0x2C59),
                    t.EUI64.convert("C6:DF:28:F9:60:33:DB:03"): t.NWK(0x1CA0),
                    # Random devices on the network
                    t.EUI64.convert("7A:BF:38:A9:59:21:A0:7A"): t.NWK(0x16B5),
                    t.EUI64.convert("10:55:FE:67:24:EA:96:D3"): t.NWK(0xBFB9),
                    t.EUI64.convert("9A:0E:10:50:00:1B:1A:5F"): t.NWK(0x1AF6),
                    t.EUI64.convert("AA:BB:CC:DD:11:22:33:44"): t.NWK(0x0ABC),
                },
                route_table={
                    t.NWK(0x16B5): t.NWK(0xBFB9),
                    t.NWK(0x1AF6): t.NWK(0x0ABC),
                },
                stack_specific={
                    "zstack": {"tclk_seed": "71e31105bb92a2d15747a0d0a042dbfd"}
                },
                metadata={"zstack": {"version": "20220102"}},
            ),
        )

    return inner


@pytest.fixture
def backup(backup_factory):
    return backup_factory()


@pytest.fixture
def z2m_backup_json():
    return {
        "metadata": {
            "format": "zigpy/open-coordinator-backup",
            "version": 1,
            "source": "zigbee-herdsman@0.13.65",
            "internal": {"date": "2021-02-08T19:35:24.761Z", "znpVersion": 2},
        },
        "stack_specific": {"zstack": {"tclk_seed": "71e31105bb92a2d15747a0d0a042dbfd"}},
        "coordinator_ieee": "932ca934d9d05d12",
        "pan_id": "9bb0",
        "extended_pan_id": "0d499199aecd3c35",
        "nwk_update_id": 18,
        "security_level": 5,
        "channel": 15,
        "channel_mask": [15, 20, 25],
        "network_key": {
            "key": "9a79d69adaec45c6f2efebafdaa307b6",
            "sequence_number": 108,
            "frame_counter": 39009277,
        },
        "devices": [
            {
                "nwk_address": "2c59",
                "ieee_address": "a31af68e199523be",
                "link_key": {
                    "key": "ca02e8bb757c94f89339d39cb3cda7be",
                    "tx_counter": 2597245184,
                    "rx_counter": 824424412,
                },
                # "is_child": True, #  Implicitly a child device
            },
            {
                "nwk_address": None,
                "ieee_address": "690c0752aad77d71",
                "link_key": {
                    "key": "857c05003e761af9689a49416a605c76",
                    "tx_counter": 3792973670,
                    "rx_counter": 1083290572,
                },
                "is_child": False,
            },
            {
                "nwk_address": None,
                "ieee_address": "a402a0dc17d817df",
                "is_child": True,
            },
            {
                "nwk_address": "1ca0",
                "ieee_address": "c6df28f96033db03",
                "is_child": True,
            },
            {
                "nwk_address": "16b5",
                "ieee_address": "7abf38a95921a07a",
                "is_child": False,
            },
            {
                "nwk_address": "bfb9",
                "ieee_address": "1055fe6724ea96d3",
                "is_child": False,
            },
            {
                "nwk_address": "1af6",
                "ieee_address": "9a0e1050001b1a5f",
                "is_child": False,
            },
            {
                "nwk_address": "abc",  # missing `0` prefix
                "ieee_address": "aabbccdd11223344",
                "is_child": False,
            },
        ],
    }


@pytest.fixture
def zigate_backup_json():
    return {
        "backup_time": "2022-07-20T17:58:16.694438+00:00",
        "network_info": {
            "extended_pan_id": "9d:ff:72:2d:19:2c:d1:01",
            "pan_id": "D08A",
            "nwk_update_id": 0,  # missing
            "nwk_manager_id": "0000",
            "channel": 15,
            "channel_mask": [15],
            "security_level": 5,
            "network_key": {
                # missing
                "key": "ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff:ff",
                "tx_counter": 0,
                "rx_counter": 0,
                "seq": 0,
                "partner_ieee": "ff:ff:ff:ff:ff:ff:ff:ff",
            },
            "tc_link_key": {
                # missing
                "key": "5a:69:67:42:65:65:41:6c:6c:69:61:6e:63:65:30:39",
                "tx_counter": 0,
                "rx_counter": 0,
                "seq": 0,
                "partner_ieee": "00:15:8d:00:06:a3:fd:fe",
            },
            "key_table": [],
            "children": [],
            "nwk_addresses": {},
            "stack_specific": {},
            "metadata": {"zigate": {"version": "3.21"}},
            "source": "zigpy-zigate@0.9.0",
        },
        "node_info": {
            "nwk": "0000",
            "ieee": "00:15:8d:00:06:a3:fd:fe",
            "logical_type": "coordinator",
        },
    }


def test_state_backup_as_dict(backup):
    obj = json.loads(json.dumps(backup.as_dict()))
    restored_backup = type(backup).from_dict(obj)
    assert backup == restored_backup


def test_state_backup_as_open_coordinator(backup):
    obj = json.loads(json.dumps(backup.as_open_coordinator_json()))
    backup2 = zigpy.backups.NetworkBackup.from_open_coordinator_json(obj)

    assert backup == backup2


def test_z2m_backup_parsing(z2m_backup_json, backup):
    backup.network_info.metadata = None
    backup.network_info.source = None
    backup.node_info.manufacturer = None
    backup.node_info.model = None
    backup.node_info.version = None
    backup.network_info.tc_link_key.tx_counter = 0
    backup.network_info.tx_power = None
    backup.network_info.route_table = {}

    for key in backup.network_info.key_table:
        key.seq = 0

    backup2 = zigpy.backups.NetworkBackup.from_open_coordinator_json(z2m_backup_json)
    backup2.network_info.metadata = None
    backup2.network_info.source = None

    # Key order may be different
    backup.network_info.key_table.sort(key=lambda k: k.key)
    backup2.network_info.key_table.sort(key=lambda k: k.key)

    assert backup == backup2


def test_from_dict_automatic(z2m_backup_json):
    backup1 = zigpy.backups.NetworkBackup.from_open_coordinator_json(z2m_backup_json)
    backup2 = zigpy.backups.NetworkBackup.from_dict(z2m_backup_json)

    assert backup1 == backup2


def test_from_dict_failure():
    with pytest.raises(ValueError):
        zigpy.backups.NetworkBackup.from_dict({"some": "json"})


def test_from_dict_future_version(backup: zigpy.backups.NetworkBackup, caplog) -> None:
    """Test that loading a backup with a future version logs a warning."""
    with caplog.at_level(logging.WARNING):
        backup = zigpy.backups.NetworkBackup.from_dict(
            {**backup.as_dict(), "version": 99}
        )

    assert "Network backup has version 99 but current backup format" in caplog.text
    assert "Downgrading is not recommended" in caplog.text

    assert backup is not None

    # Verify version was downgraded to current BACKUP_FORMAT_VERSION
    assert backup.version == zigpy.backups.BACKUP_FORMAT_VERSION


def test_backup_compatibility(backup_factory):
    backup1 = backup_factory()
    assert backup1.is_compatible_with(backup1)

    # Incompatible due to different coordinator IEEE
    backup2 = backup_factory()
    backup2.node_info.ieee = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")
    assert not backup2.supersedes(backup1)
    assert not backup1.supersedes(backup2)
    assert not backup1.is_compatible_with(backup2)

    # NWK frame counter must always be greater
    backup3 = backup_factory()
    backup3.network_info.network_key.tx_counter -= 1
    assert backup3.is_compatible_with(backup1)
    assert not backup3.supersedes(backup1)

    backup4 = backup_factory()
    backup4.network_info.network_key.tx_counter += 1
    assert backup4.is_compatible_with(backup1)
    assert backup4.supersedes(backup1)


async def test_backup_completeness(backup, zigate_backup_json):
    assert backup.is_complete()

    zigate_backup = zigpy.backups.NetworkBackup.from_dict(zigate_backup_json)
    assert not zigate_backup.is_complete()

    backups = zigpy.backups.BackupManager(None)

    with pytest.raises(ValueError):
        await backups.restore_backup(zigate_backup)


async def test_add_backup(backup_factory):
    backups = zigpy.backups.BackupManager(None)

    # First backup
    backup1 = backup_factory()
    backups.add_backup(backup1)
    assert backups.backups == [backup1]

    # Adding the same backup twice will do nothing
    backups.add_backup(backup1)
    assert backups.backups == [backup1]

    # Adding an identical backup that is newer replaces the old one
    backup2 = backup_factory()
    backup2.backup_time += timedelta(hours=1)
    backups.add_backup(backup2)
    assert backups.backups == [backup2]

    # An even more recent one with a rolled back frame counter is appended
    backup3 = backup_factory()
    backup3.backup_time += timedelta(hours=2)
    backup3.network_info.network_key.tx_counter -= 1000
    backups.add_backup(backup3)
    assert backups.backups == [backup2, backup3]

    # A final one replacing them both is added
    backup4 = backup_factory()
    backup4.backup_time += timedelta(hours=3)
    backup4.network_info.network_key.tx_counter += 1000
    backups.add_backup(backup4)
    assert backups.backups == [backup4]

    # An incompatible backup will be added to the list. Nothing will be replaced.
    backup5 = backup_factory()
    backup5.network_info.pan_id += 1
    backups.add_backup(backup5)
    assert backups.backups == [backup4, backup5]


async def test_restore_backup_create_new(app, backup):
    backups = zigpy.backups.BackupManager(app)
    backups.create_backup = AsyncMock()

    await backups.restore_backup(backup)
    app.write_network_info.assert_called_once()
    backups.create_backup.assert_called_once()

    app.write_network_info.reset_mock()
    backups.create_backup.reset_mock()

    await backups.restore_backup(backup, create_new=False)
    app.write_network_info.assert_called_once()
    backups.create_backup.assert_not_called()  # Won't be called



"""Primitive data structures."""

from __future__ import annotations

import asyncio
import bisect
import collections
from collections.abc import Coroutine
import contextlib
from fractions import Fraction
import functools
import heapq
import logging
import math
import types
import typing
from typing import Any, Generic, Self, TypeVar
import warnings

_LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T", bound=contextlib.AbstractAsyncContextManager)


class WrappedContextManager(Generic[_T]):
    def __init__(
        self,
        context_manager: _T,
        on_enter: typing.Callable[[], Coroutine[Any, Any, Any]],
    ) -> None:
        self.on_enter = on_enter
        self.context_manager = context_manager

    async def __aenter__(self) -> _T:
        await self.on_enter()
        return self.context_manager

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        await self.context_manager.__aexit__(exc_type, exc, traceback)


class PriorityDynamicBoundedSemaphore:
    """`asyncio.BoundedSemaphore` with public interface to change the max value."""

    def __init__(self, value: int = 0) -> None:
        self._value: int = value
        self._max_value: int = value
        self._comparison_counter: int = 0
        self._waiters: list[tuple[int, int, asyncio.Future]] = []
        self._loop: asyncio.AbstractEventLoop | None = None

    def _get_loop(self) -> asyncio.AbstractEventLoop:
        loop = asyncio.get_running_loop()

        if self._loop is None:
            self._loop = loop

        if loop is not self._loop:
            raise RuntimeError(f"{self!r} is bound to a different event loop")

        return loop

    def _wake_up_next(self) -> bool:
        """Wake up the first waiter that isn't done."""
        if not self._waiters:
            return False

        for _, _, fut in self._waiters:
            if not fut.done():
                self._value -= 1
                fut.set_result(True)
                # `fut` is now `done()` and not `cancelled()`.
                return True
        return False

    def cancel_waiting(self, exc: BaseException) -> None:
        """Cancel all waiters with the given exception."""
        for _, _, fut in self._waiters:
            if not fut.done():
                fut.set_exception(exc)

    @property
    def value(self) -> int:
        return self._value

    @property
    def max_value(self) -> int:
        return self._max_value

    @max_value.setter
    def max_value(self, new_value: int) -> None:
        """Update the semaphore's max value."""
        if new_value < 0:
            raise ValueError(f"Semaphore value must be >= 0: {new_value!r}")

        delta = new_value - self._max_value
        self._value += delta
        self._max_value += delta

        # Wake up any pending waiters
        for _ in range(max(0, delta)):
            if not self._wake_up_next():
                break

    @property
    def num_waiting(self) -> int:
        return len(self._waiters)

    def locked(self) -> bool:
        """Returns True if semaphore cannot be acquired immediately."""
        # Due to state, or FIFO rules (must allow others to run first).
        return self._value <= 0 or (any(not w.cancelled() for _, _, w in self._waiters))

    async def acquire(self, priority: int = 0) -> typing.Literal[True]:
        """Acquire a semaphore.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately.  If it is
        zero on entry, block, waiting until some other task has
        called release() to make it larger than 0, and then return
        True.
        """
        if not self.locked():
            # Maintain FIFO, wait for others to start even if _value > 0.
            self._value -= 1
            return True

        # To ensure that our objects don't have to be themselves comparable, we
        # maintain a global count and increment it on every insert. This way,
        # the tuple `(-priority, count, item)` will never have to compare `item`.
        self._comparison_counter += 1

        fut = self._get_loop().create_future()
        obj = (-priority, self._comparison_counter, fut)
        bisect.insort_right(self._waiters, obj)

        try:
            try:
                await fut
            finally:
                self._waiters.remove(obj)
        except asyncio.CancelledError:
            # Only undo the bookkeeping if the slot was actually granted. If the future
            # instead carries an exception (e.g. from `cancel_waiting`), no slot was
            # ever acquired and there is nothing to release.
            if fut.done() and not fut.cancelled() and fut.exception() is None:
                # Our Future was successfully set to True via _wake_up_next(),
                # but we are not about to successfully acquire(). Therefore we
                # must undo the bookkeeping already done and attempt to wake
                # up someone else.
                self._value += 1
            raise

        finally:
            # New waiters may have arrived but had to wait due to FIFO.
            # Wake up as many as are allowed.
            while self._value > 0:
                if not self._wake_up_next():
                    break  # There was no-one to wake up.
        return True

    def release(self) -> None:
        """Release a semaphore, incrementing the internal counter by one.

        When it was zero on entry and another task is waiting for it to
        become larger than zero again, wake up that task.
        """
        if self._value >= self._max_value:
            raise ValueError("Semaphore released too many times")

        self._value += 1
        self._wake_up_next()

    def __call__(self, priority: int = 0) -> WrappedContextManager[Self]:
        """Allows specifying the priority by calling the context manager.

        This allows both `async with sem:` and `async with sem(priority=5):`.
        """
        return WrappedContextManager(
            context_manager=self,
            on_enter=lambda: self.acquire(priority),
        )

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.release()

    def __repr__(self) -> str:
        if self.locked():
            extra = f"locked, max value:{self._max_value}, waiters:{len(self._waiters)}"
        else:
            extra = f"unlocked, value:{self._value}, max value:{self._max_value}"

        return f"<{self.__class__.__name__} [{extra}]>"


class PriorityLock(PriorityDynamicBoundedSemaphore):
    def __init__(self) -> None:
        super().__init__(value=1)

    @PriorityDynamicBoundedSemaphore.max_value.setter
    def max_value(self, new_value: int) -> None:
        """Update the locks's max value."""
        raise ValueError("Max value of lock cannot be updated")


# Backwards compatibility
DynamicBoundedSemaphore = PriorityDynamicBoundedSemaphore


class ReschedulableTimeout:
    """Timeout object made to be efficiently rescheduled continuously."""

    def __init__(self, callback: typing.Callable[[], None]) -> None:
        self._timer: asyncio.TimerHandle | None = None
        self._callback = callback

        self._when: float = 0

    @functools.cached_property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def _timeout_trigger(self) -> None:
        now = self._loop.time()

        # If we triggered early, reschedule
        if self._when > now:
            self._reschedule()
            return

        self._timer = None
        self._callback()

    def _reschedule(self) -> None:
        if self._timer is not None:
            self._timer.cancel()

        self._timer = self._loop.call_at(self._when, self._timeout_trigger)

    def reschedule(self, delay: float) -> None:
        self._when = self._loop.time() + delay

        # If the current timer will expire too late (or isn't running), reschedule
        if self._timer is None or self._timer.when() > self._when:
            self._reschedule()

    def cancel(self) -> None:
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None


class Debouncer:
    """Generic debouncer supporting per-invocation expiration."""

    def __init__(self) -> None:
        self._times: dict[typing.Any, float] = {}
        self._queue: list[tuple[float, int, typing.Any]] = []

        self._last_time: int = 0
        self._dedup_counter: int = 0

    @functools.cached_property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def clean(self, now: float | None = None) -> None:
        """Clean up stale timers."""
        if now is None:
            now = self._loop.time()

        # We store the negative expiration time to ensure we can pop expiring objects
        while self._queue and -self._queue[-1][0] < now:
            _, _, obj = self._queue.pop()
            self._times.pop(obj)

    def is_filtered(self, obj: typing.Any, now: float | None = None) -> bool:
        """Check if an object will be filtered."""
        if now is None:
            now = self._loop.time()

        # Clean up stale timers
        self.clean(now)

        # If an object still exists after cleaning, it won't be expired
        return obj in self._times

    def filter(self, obj: typing.Any, expire_in: float) -> bool:
        """Check if an object should be filtered. If not, store it."""
        now = self._loop.time()

        # For platforms with low-resolution clocks, we need to make sure that `obj` will
        # never be compared by `heapq`!
        if now > self._last_time:
            self._last_time = now
            self._dedup_counter = 0

        self._dedup_counter += 1

        # If the object is filtered, do nothing
        if self.is_filtered(obj, now=now):
            return True

        # Otherwise, queue it
        self._times[obj] = now + expire_in
        bisect.insort_right(self._queue, (-(now + expire_in), self._dedup_counter, obj))

        return False

    def __repr__(self) -> str:
        """String representation of the debouncer."""
        return f"<{self.__class__.__name__} [tracked:{len(self._queue)}]>"


class _LimiterContext:
    """Helper class to manage the async context for the RequestLimiter."""

    def __init__(self, limiter: RequestLimiter, priority: int) -> None:
        self._limiter = limiter
        self._priority = priority

    async def __aenter__(self) -> None:
        """Acquire a slot from the limiter."""
        await self._limiter._acquire(self._priority)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        """Release the slot back to the limiter."""
        self._limiter._release(self._priority)


class RequestLimiter:
    """Limits concurrent requests with cascading capacity for multiple priority levels."""

    def __init__(self, max_concurrency: int, capacities: dict[int, float]) -> None:
        """Initializes the RequestLimiter."""
        if max_concurrency < 0:
            raise ValueError(f"max_concurrency must be >= 0: {max_concurrency}")

        self._lock = asyncio.Lock()
        self._capacity_fractions = capacities
        self._sorted_priorities = sorted(capacities.keys())

        self._cumulative_capacity: dict[int, int] = {}
        self._max_concurrency = max_concurrency
        self._recalculate_capacity()

        self._active_requests_by_tier: typing.Counter[int] = collections.Counter()
        self._waiters: list[tuple[int, int, asyncio.Future]] = []
        self._comparison_counter = 0

    @property
    def active_requests(self) -> int:
        """Returns the total number of currently running requests."""
        return sum(self._active_requests_by_tier.values())

    @property
    def waiting_requests(self) -> int:
        """Returns the number of requests waiting for a slot."""
        return len(self._waiters)

    @property
    def max_concurrency(self) -> int:
        """Returns the maximum concurrency of the limiter."""
        return self._max_concurrency

    @max_concurrency.setter
    def max_concurrency(self, new_value: int) -> None:
        """Updates the maximum concurrency of the limiter."""
        self._max_concurrency = new_value
        self._recalculate_capacity()
        self._wake_waiters()

    @property
    def max_value(self) -> int:
        """Deprecated alias for `max_concurrency`."""
        warnings.warn(
            "`max_value` is deprecated, use `max_concurrency` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.max_concurrency

    @max_value.setter
    def max_value(self, new_value: int) -> None:
        """Deprecated setter alias for `max_concurrency`."""
        warnings.warn(
            "`max_value` is deprecated, use `max_concurrency` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.max_concurrency = new_value

    def _recalculate_capacity(self) -> None:
        # Assume that all of the fractions are simple
        divisors = [
            Fraction.from_float(f).limit_denominator(10).denominator
            for f in self._capacity_fractions.values()
        ]

        lcm = math.lcm(*divisors)

        if self._max_concurrency % lcm != 0:
            next_best = lcm - self._max_concurrency % lcm
            assert next_best > 0

            _LOGGER.warning(
                "Requested adapter concurrency %d is not compatible with priority fractions %r. Increasing concurrency to %d.",
                self._max_concurrency,
                self._capacity_fractions,
                self._max_concurrency + next_best,
            )
            self._max_concurrency += next_best

        cumulative_capacity = 0

        for priority in self._sorted_priorities:
            portion = self._capacity_fractions[priority] * self._max_concurrency
            cumulative_capacity += round(portion)

            self._cumulative_capacity[priority] = cumulative_capacity

    def __call__(self, priority: int = 0) -> _LimiterContext:
        """Returns an async context manager to safely acquire and release a slot."""
        return _LimiterContext(self, priority)

    def _get_effective_priority_tier(self, priority: int) -> int:
        """Finds the capacity tier that the given priority falls into."""
        if priority < self._sorted_priorities[0]:
            raise ValueError(
                f"Priority {priority} is lower than the lowest known priority "
                f"{self._sorted_priorities[0]} and has no allocated capacity"
            )

        idx = bisect.bisect_right(self._sorted_priorities, priority)
        return self._sorted_priorities[idx - 1]

    def locked(self, priority: int) -> bool:
        """Checks if a request with a given priority can run."""
        effective_tier = self._get_effective_priority_tier(priority)
        limit = self._cumulative_capacity[effective_tier]
        waiting_requests = 0

        if limit == 0:
            return True

        for tier, count in self._active_requests_by_tier.items():
            if tier > effective_tier:
                continue

            waiting_requests += count
            if waiting_requests >= limit:
                return True

        return False

    def _wake_waiters(self) -> None:
        """Wakes up any waiting tasks that can now run."""
        while self._waiters:
            waiter_priority, _, fut = self._waiters[0]
            priority = -waiter_priority  # We flip the sign when storing, for comparison

            if not self.locked(priority):
                heapq.heappop(self._waiters)
                if not fut.done():
                    effective_tier = self._get_effective_priority_tier(priority)
                    self._active_requests_by_tier[effective_tier] += 1
                    fut.set_result(None)
            else:
                break

    async def _acquire(self, priority: int = 0) -> bool:
        """Acquires a slot in the limiter, waiting if necessary."""
        effective_tier = self._get_effective_priority_tier(priority)

        # A task can run immediately if it has capacity AND it has a higher
        # priority than any task already waiting. This allows high-priority
        # tasks to jump the queue, while maintaining FIFO for tasks of the
        # same priority.
        highest_waiter_priority = (
            -self._waiters[0][0] if self._waiters else -float("inf")
        )

        if not self.locked(priority) and priority > highest_waiter_priority:
            self._active_requests_by_tier[effective_tier] += 1
            return True

        # To ensure that our objects don't have to be themselves comparable, we
        # maintain a global count and increment it on every insert. This way,
        # the tuple `(-priority, count, item)` will never have to compare `item`.
        self._comparison_counter += 1
        fut = asyncio.get_running_loop().create_future()
        waiter_obj = (-priority, self._comparison_counter, fut)
        heapq.heappush(self._waiters, waiter_obj)

        try:
            try:
                await fut
                return True
            finally:
                if waiter_obj in self._waiters:
                    self._waiters.remove(waiter_obj)
                    heapq.heapify(self._waiters)
        except asyncio.CancelledError:
            # Only undo the bookkeeping if the slot was actually granted. If the future
            # instead carries an exception (e.g. from `cancel_waiting`), no slot was
            # ever acquired and there is nothing to release.
            if fut.done() and not fut.cancelled() and fut.exception() is None:
                self._active_requests_by_tier[effective_tier] -= 1

            raise
        finally:
            self._wake_waiters()

    def _release(self, priority: int = 0) -> None:
        """Releases an acquired slot back to the limiter."""
        effective_tier = self._get_effective_priority_tier(priority)
        assert self._active_requests_by_tier[effective_tier] > 0
        self._active_requests_by_tier[effective_tier] -= 1
        self._wake_waiters()

    def cancel_waiting(self, exc: BaseException) -> None:
        """Cancel all waiters with the given exception."""
        for _, _, fut in self._waiters:
            if not fut.done():
                fut.set_exception(exc)

    def __repr__(self) -> str:
        """Provides a string representation of the limiter's state."""
        return (
            f"<{self.__class__.__name__}("
            f"max_concurrency={self._max_concurrency}"
            f", active={self.active_requests}"
            f", waiting={self.waiting_requests}"
            f")>"
        )



"""Zigpy Constants."""

from __future__ import annotations

SIG_ENDPOINTS = "endpoints"
SIG_EP_INPUT = "input_clusters"
SIG_EP_OUTPUT = "output_clusters"
SIG_EP_PROFILE = "profile_id"
SIG_EP_TYPE = "device_type"
SIG_MANUFACTURER = "manufacturer"
SIG_MODEL = "model"
SIG_MODELS_INFO = "models_info"
SIG_NODE_DESC = "node_desc"
SIG_SKIP_CONFIG = "skip_configuration"

INTERFERENCE_MESSAGE = (
    "If you are having problems joining new devices, are missing sensor"
    " updates, or have issues keeping devices joined, ensure your"
    " coordinator is away from interference sources such as USB 3.0"
    " devices, SSDs, WiFi routers, etc."
)

APS_REPLY_TIMEOUT = 5
APS_REPLY_TIMEOUT_EXTENDED = 28

QUIRKS_REPO_URL = "https://github.com/zigpy/zha-device-handlers"



from __future__ import annotations

import asyncio
import logging
import time
import typing

import pytest

from zigpy import util
from zigpy.types.named import EUI64, KeyData

from .async_mock import AsyncMock, MagicMock, call, patch, sentinel


class Listenable(util.ListenableMixin):
    def __init__(self):
        self._listeners = {}


def test_listenable():
    listen = Listenable()

    # Python 3.7 guarantees dict ordering so this will be called first to test error
    # handling
    broken_listener = MagicMock()
    broken_listener.event.side_effect = Exception()
    listen.add_listener(broken_listener)

    listener = MagicMock(spec_set=["event"])
    listen.add_listener(listener)
    listen.add_listener(listener)

    context_listener = MagicMock(spec_set=["event"])
    listen.add_context_listener(context_listener)

    listen.listener_event("event", "test1")
    listener.event.assert_has_calls([call("test1"), call("test1")], any_order=True)
    context_listener.event.assert_has_calls([call(listen, "test1")], any_order=True)
    broken_listener.event.assert_has_calls([call("test1")], any_order=True)
    assert listener.event.call_count == 2
    assert context_listener.event.call_count == 1
    assert broken_listener.event.call_count == 1

    listen.listener_event("non_existing_event", "test2")
    listener.event.assert_has_calls([call("test1"), call("test1")], any_order=True)
    context_listener.event.assert_has_calls([call(listen, "test1")], any_order=True)
    broken_listener.event.assert_has_calls([call("test1")], any_order=True)
    assert listener.event.call_count == 2
    assert context_listener.event.call_count == 1
    assert broken_listener.event.call_count == 1

    listen.remove_listener(object())
    listen.remove_listener(listener)
    listen.remove_listener(listener)
    listen.remove_listener(broken_listener)
    listen.remove_listener(context_listener)
    listen.listener_event("event", "test1")
    assert listener.event.call_count == 2
    assert context_listener.event.call_count == 1
    assert broken_listener.event.call_count == 1


def test_listenable_mutating():
    listen = Listenable()
    mutating_listener = MagicMock()
    mutating_listener.event.side_effect = lambda value: listen.remove_listener(
        mutating_listener
    )

    listen.add_listener(mutating_listener)
    listen.listener_event("event", "value")

    assert mutating_listener.event.mock_calls == [call("value")]
    assert listen._listeners == {}


class Logger(util.LocalLogMixin):
    log = MagicMock()


def test_log():
    log = Logger()
    log.debug("Test debug")
    log.exception("Test exception")
    log.info("Test info")
    log.warning("Test warn")
    log.error("Test error")


def test_log_stacklevel():
    class MockHandler(logging.Handler):
        emit = MagicMock()

    handler = MockHandler()

    LOGGER = logging.getLogger("test_log_stacklevel")
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(handler)

    class TestClass(util.LocalLogMixin):
        def log(self, lvl, msg, *args, **kwargs):
            LOGGER.log(lvl, msg, *args, **kwargs)

        def test_method(self):
            self.info("Test1")
            LOGGER.info("Test2")

    TestClass().test_method()

    assert handler.emit.call_count == 2

    indirect_call, direct_call = handler.emit.mock_calls
    (indirect,) = indirect_call.args
    (direct,) = direct_call.args

    assert indirect.message == "Test1"
    assert direct.message == "Test2"
    assert direct.levelname == indirect.levelname

    assert direct.module == indirect.module
    assert direct.filename == indirect.filename
    assert direct.funcName == indirect.funcName
    assert direct.lineno == indirect.lineno + 1


async def _test_retry(exception, retry_exceptions, n):
    counter = 0

    async def count():
        nonlocal counter
        counter += 1
        if counter <= n:
            exc = exception()
            exc._counter = counter
            raise exc

    await util.retry(count, retry_exceptions)
    return counter


async def test_retry_no_retries():
    counter = await _test_retry(Exception, Exception, 0)
    assert counter == 1


async def test_retry_always():
    with pytest.raises(ValueError) as exc_info:
        await _test_retry(ValueError, (IndexError, ValueError), 999)
    assert exc_info.value._counter == 3


async def test_retry_once():
    counter = await _test_retry(ValueError, ValueError, 1)
    assert counter == 2


async def _test_retryable(exception, retry_exceptions, n, tries=3, delay=0.001):
    counter = 0

    @util.retryable(retry_exceptions, tries=tries, delay=delay)
    async def count(x, y, z):
        assert x == y == z == 9
        nonlocal counter
        counter += 1
        if counter <= n:
            exc = exception()
            exc._counter = counter
            raise exc

    await count(9, 9, 9)
    return counter


async def test_retryable_no_retry():
    counter = await _test_retryable(Exception, Exception, 0, 0, 0)
    assert counter == 1


async def test_retryable_exception_no_retry():
    with pytest.raises(Exception) as exc_info:
        await _test_retryable(Exception, Exception, 1, 0, 0)
    assert exc_info.value._counter == 1


async def test_retryable_no_retries():
    counter = await _test_retryable(Exception, Exception, 0)
    assert counter == 1


async def test_retryable_always():
    with pytest.raises(ValueError) as exc_info:
        await _test_retryable(ValueError, (IndexError, ValueError), 999)
    assert exc_info.value._counter == 3


async def test_retryable_once():
    counter = await _test_retryable(ValueError, ValueError, 1)
    assert counter == 2


def test_zigbee_security_hash():
    message = bytes.fromhex("11223344556677884AF7")
    key = util.aes_mmo_hash(message)
    assert key == KeyData.convert("41618FC0C83B0E14A589954B16E31466")

    message = bytes.fromhex("7A939723A5C639B269161802819B")
    key = util.aes_mmo_hash(message)
    assert key == KeyData.convert("F93903721685FD329D26849B90F2959A")

    message = bytes.fromhex("83FED3407A939723A5C639B269161802AEBB")
    key = util.aes_mmo_hash(message)
    assert key == KeyData.convert("333C23686079468EB27BA24BD9C7E564")


@pytest.mark.parametrize(
    ("message", "expected_key"),
    [
        (
            bytes.fromhex("11223344556677884AF7"),
            KeyData.convert("41618FC0C83B0E14A589954B16E31466"),
        ),
        (
            bytes.fromhex("83FED3407A939723A5C639B26916D505C3B5"),
            KeyData.convert("66B6900981E1EE3CA4206B6B861C02BB"),
        ),
    ],
)
def test_convert_install_code(message, expected_key):
    key = util.convert_install_code(message)
    assert key == expected_key


def test_fail_convert_install_code():
    with pytest.raises(ValueError, match="Invalid install code length"):
        util.convert_install_code(b"")

    with pytest.raises(ValueError, match="Invalid install code CRC"):
        util.convert_install_code(b"\x11\x22\x33\x44\x55\x66\x77\x88\xff\xff")


async def test_async_listener():
    listenable = Listenable()

    listener_1 = MagicMock(spec=["async_event"])
    listener_1.async_event.side_effect = AsyncMock(return_value=sentinel.result_1)

    listener_2 = MagicMock(spec=["async_event"])
    listener_2.async_event.side_effect = AsyncMock(return_value=sentinel.result_2)

    failed = MagicMock(spec=["async_event"])
    failed.async_event = AsyncMock(side_effect=RuntimeError("async listener exception"))

    listenable.add_listener(listener_1)
    listenable.add_context_listener(listener_2)
    listenable.add_listener(failed)

    r = await listenable.async_event("async_event", sentinel.data)
    assert len(r) == 2
    assert sentinel.result_1 in r
    assert sentinel.result_2 in r

    assert listener_1.async_event.call_count == 1
    assert listener_1.async_event.call_args[0][0] is sentinel.data

    # context listener
    assert listener_2.async_event.call_count == 1
    assert listener_2.async_event.call_args[0][0] is listenable
    assert listener_2.async_event.call_args[0][1] is sentinel.data

    # failed listener
    assert failed.async_event.call_count == 1
    assert failed.async_event.call_args[0][0] is sentinel.data

    r = await listenable.async_event("no_such_event", sentinel.no_data)
    assert r == []


class _ClusterMock(util.CatchingTaskMixin):
    """Test class."""

    def __init__(self, logger):
        logger.setLevel(logging.DEBUG)
        self._logger = logger

    def log(self, lvl, msg, *args, **kwargs):
        return self._logger.log(lvl, msg, *args, **kwargs)

    async def a(self, exception=None):
        self.debug("test a")
        return await self._b(exception)

    async def _b(self, exception):
        self.warning("test b")
        if exception is None:
            return True
        raise exception()


@patch("zigpy.util.CatchingTaskMixin.catching_coro")
async def test_create_catching_task(catching_coro_mock):
    """Test catching task."""
    mock_cluster = _ClusterMock(logging.getLogger(__name__))
    coro = AsyncMock()
    mock_cluster.create_catching_task(coro)
    assert catching_coro_mock.call_count == 1
    assert catching_coro_mock.call_args[0][0] is coro


async def test_catching_coro(caplog):
    """Test catching_coro no exception."""
    caplog.set_level(level=logging.DEBUG)
    mock_cluster = _ClusterMock(logging.getLogger(__name__))
    await mock_cluster.catching_coro(mock_cluster.a())

    records = [r for r in caplog.records if r.name == __name__]
    assert records[0].levelno == logging.DEBUG
    assert records[0].message == "test a"
    assert records[1].levelno == logging.WARNING
    assert records[1].message == "test b"
    assert len(records) == 2


@pytest.mark.parametrize("exception", [None, asyncio.TimeoutError])
async def test_catching_task_expected_exception(exception, caplog):
    """Test CatchingTaskMixin allowed exceptions."""
    mock_cluster = _ClusterMock(logging.getLogger("expected_exceptions"))
    await mock_cluster.catching_coro(
        mock_cluster.a(asyncio.TimeoutError), exceptions=exception
    )

    records = [r for r in caplog.records if r.name == "expected_exceptions"]
    assert records[0].levelno == logging.DEBUG
    assert records[0].message == "test a"
    assert records[1].levelno == logging.WARNING
    assert records[1].message == "test b"
    assert len(records) == 2


@pytest.mark.parametrize(
    ("to_raise", "exception"),
    [(RuntimeError, None), (asyncio.TimeoutError, RuntimeError)],
)
async def test_catching_task_unexpected_exception(to_raise, exception, caplog):
    """Test CatchingTaskMixin unexpected exceptions."""
    mock_cluster = _ClusterMock(logging.getLogger("unexpected_exceptions"))
    await mock_cluster.catching_coro(mock_cluster.a(to_raise), exceptions=exception)

    records = [r for r in caplog.records if r.name == "unexpected_exceptions"]
    assert records[0].levelno == logging.DEBUG
    assert records[0].message == "test a"
    assert records[1].levelno == logging.WARNING
    assert records[1].message == "test b"
    assert records[2].levelno == logging.ERROR
    assert records[2].message.startswith("Traceback (most recent call last)")
    assert len(records) == 3


@pytest.mark.parametrize(
    "plot",
    [
        r"""
            11  #########
            12  #################################################################
            13  ##################################################################
            14  ################################################################
            15  ##############
            16  ######################################################
            17  #########################
            18  ##########
            19  ###########
            20  ################
            21  ###########
            22  ###########
            23  #
            24  ##
           [25] #################
            26  ##
        """,
        r"""
            11  ##################
            12
            13
            14  ########
            15  ##################
            16  ########
            17
            18
            19  ########
            20  ##################
            21  ########
            22
            23
            24  ########
           [25] ##################
            26
        """,
        r"""
            11  ##################
            12  #
            13  #
            14  ########
            15  ##################
            16  ########
            17
            18
            19  ########
           [20] ##################
            21  ########
            22
            23
            24  ########
            25  ##################
            26  ########
        """,
    ],
)
def test_picking_optimal_channel(plot):
    expected_channel = int(plot.split("[")[1].split("]")[0])
    plot = plot.replace("[", " ").replace("]", " ")
    channel_energy = {
        int(line.split()[0]): line.count("#") for line in plot.strip().splitlines()
    }

    assert util.pick_optimal_channel(channel_energy) == expected_channel


@pytest.mark.parametrize(
    ("input_relays", "expected_relays"),
    [
        ([0x0000, 0x0000, 0x0001, 0x0001, 0x0002], [0x0001, 0x0002]),
        ([0x0001, 0x0002], [0x0001, 0x0002]),
        ([], []),
        ([0x0000], []),
    ],
)
def test_relay_filtering(input_relays: list[int], expected_relays: list[int]):
    assert util.filter_relays(input_relays) == expected_relays


async def test_combine_concurrent_calls():
    class TestFuncs:
        def __init__(self):
            self.slow_calls = 0
            self.slow_error_calls = 0

        async def slow(self, n=None):
            await asyncio.sleep(0.1)
            self.slow_calls += 1
            return (self.slow_calls, n)

        async def slow_error(self, n=None):
            await asyncio.sleep(0.1)
            self.slow_error_calls += 1
            raise RuntimeError

        combined_slow = util.combine_concurrent_calls(slow)
        combined_slow_error = util.combine_concurrent_calls(slow_error)

    f = TestFuncs()

    assert f.slow_calls == 0

    await f.slow()
    assert f.slow_calls == 1

    await f.combined_slow()
    assert f.slow_calls == 2

    results = await asyncio.gather(*[f.combined_slow() for _ in range(5)])
    assert results == [(3, None)] * 5
    assert f.slow_calls == 3

    results = await asyncio.gather(*[f.combined_slow() for _ in range(5)])
    assert results == [(4, None)] * 5
    assert f.slow_calls == 4

    # Unique keyword arguments
    results = await asyncio.gather(*[f.combined_slow(n=i) for i in range(5)])
    assert results == [(5 + i, 0 + i) for i in range(5)]
    assert f.slow_calls == 9

    # Non-unique keyword arguments
    results = await asyncio.gather(*[f.combined_slow(i // 2) for i in range(5)])
    assert results == [(10, 0), (10, 0), (11, 1), (11, 1), (12, 2)]
    assert f.slow_calls == 12

    # Mixed keyword and non-keyword
    results = await asyncio.gather(
        f.combined_slow(0),
        f.combined_slow(n=0),
        f.combined_slow(1),
        f.combined_slow(n=1),
        f.combined_slow(n=1),
    )
    assert results == [(13, 0), (13, 0), (14, 1), (14, 1), (14, 1)]
    assert f.slow_calls == 14

    assert f.slow_error_calls == 0

    with pytest.raises(RuntimeError):
        await f.slow_error()

    assert f.slow_error_calls == 1

    for coro in asyncio.as_completed([f.combined_slow_error() for _ in range(5)]):
        with pytest.raises(RuntimeError):
            await coro

    assert f.slow_error_calls == 2


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_deprecated():
    @util.deprecated("This function is deprecated")
    def foo():
        return 1

    with pytest.deprecated_call():
        foo()

    class Bar:
        pass

    obj = util.deprecated_attrs({"foo": Bar})

    assert obj("foo") == Bar

    with pytest.raises(AttributeError):
        obj("baz")


async def test_async_iterate_in_chunks() -> None:
    def iterator(n: int) -> typing.Generator[int, None, None]:
        for i in range(n):
            time.sleep(0.1)
            yield i

    chunks = [c async for c in util.async_iterate_in_chunks(iterator(10), chunk_size=3)]
    assert chunks == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]


@pytest.mark.parametrize(
    ("code", "ieee", "link_key"),
    [
        # Atlantic Galapagos electric heater (core#155014)
        (
            "Z:22B78EFEFF16A720$I:B1817AFD64D8BB3A8F275C2117327AEC684B",
            EUI64.convert("20:A7:16:FF:FE:8E:B7:22"),  # reversed!
            util.convert_install_code(
                bytes.fromhex("B1817AFD64D8BB3A8F275C2117327AEC684B")
            ),
        ),
        # Inovelli, others
        (
            "Z:943469FFFE05CE55$I:FC5ECD1E8DFD3E0603E3689F8D0226BBF80B",
            EUI64.convert("94:34:69:FF:FE:05:CE:55"),
            util.convert_install_code(
                bytes.fromhex("FC5ECD1E8DFD3E0603E3689F8D0226BBF80B")
            ),
        ),
        # Muller-Light, Schneider, others
        (
            "000D6FFFFE63CCA0|58A4DD6123F38ECF22B17FDA0D95A43DED19",
            EUI64.convert("00:0D:6F:FF:FE:63:CC:A0"),
            util.convert_install_code(
                bytes.fromhex("58A4DD6123F38ECF22B17FDA0D95A43DED19")
            ),
        ),
        # Aqara
        (
            "G$M:69775$S:680S00003915$D:0000000017B2335C%Z$A:54EF44100006E7DF$I:3313A005E177A647FC7925620AB207C4BEF5",
            EUI64.convert("54:EF:44:10:00:06:E7:DF"),
            util.convert_install_code(
                bytes.fromhex("3313A005E177A647FC7925620AB207C4BEF5")
            ),
        ),
        # Somfy?
        (
            "Z:4CC206FFFE805E7A$I:560BCC1FF50E704B9BA93A5CA817C35E799F$D:202%B:4CC206805E7A$P:613727%M:1220$F:0024",
            EUI64.convert("4C:C2:06:FF:FE:80:5E:7A"),
            util.convert_install_code(
                bytes.fromhex("560BCC1FF50E704B9BA93A5CA817C35E799F")
            ),
        ),
        # Hue
        (
            "HUE:Z:2E5D2E5F401B82F1A45621A50B551F726101 M:001788010CE5C843 D:H2504 A:2266",
            EUI64.convert("00:17:88:01:0C:E5:C8:43"),
            util.convert_install_code(
                bytes.fromhex("2E5D2E5F401B82F1A45621A50B551F726101")
            ),
        ),
        # Bosch (install code)
        (
            "RB01SG0D8310182648007000000000000000000094DEB8FFFE41F6D6DLKA87710C3E5C332E5327EE532C3C310E5DFE0",
            EUI64.convert("94:DE:B8:FF:FE:41:F6:D6"),
            util.convert_install_code(
                bytes.fromhex("A87710C3E5C332E5327EE532C3C310E5DFE0")
            ),
        ),
        # Bosch (link key)
        (
            "RB01SG0D836591B3CC0010000000000000000000000D6F0017E0870CDLK999F98A7DFBCA6DD3955823AD9089631",
            EUI64.convert("00:0D:6F:00:17:E0:87:0C"),
            KeyData.convert("999F98A7DFBCA6DD3955823AD9089631"),
        ),
    ],
)
def test_qr_code_parsing(code: str, ieee: EUI64, link_key: KeyData) -> None:
    """Test install code QR parsing."""
    parsed_ieee, parsed_link_key = util.parse_install_code_qr(code)
    assert parsed_ieee == ieee
    assert parsed_link_key == link_key


def test_qr_code_parsing_failure() -> None:
    """Test install code QR parsing failure."""
    with pytest.raises(ValueError, match="Unknown QR code format: "):
        util.parse_install_code_qr("some invalid QR code")



"""Topology builder."""

from __future__ import annotations

import asyncio
import collections
import itertools
import logging
import random
import typing

import zigpy.config
import zigpy.device
import zigpy.types as t
import zigpy.util
import zigpy.zdo.types as zdo_t

LOGGER = logging.getLogger(__name__)
REQUEST_DELAY = (1.0, 1.5)

if typing.TYPE_CHECKING:
    import zigpy.application


class ScanNotSupported(Exception):
    pass


INVALID_NEIGHBOR_IEEES = {
    t.EUI64.convert("00:00:00:00:00:00:00:00"),
    t.EUI64.convert("ff:ff:ff:ff:ff:ff:ff:ff"),
}


class Topology(zigpy.util.ListenableMixin):
    """Topology scanner."""

    def __init__(self, app: zigpy.application.ControllerApplication) -> None:
        """Instantiate."""
        self._app: zigpy.application.ControllerApplication = app
        self._listeners: dict = {}
        self._scan_task: asyncio.Task | None = None
        self._scan_loop_task: asyncio.Task | None = None

        # Keep track of devices that do not support scanning
        self._neighbors_unsupported: set[t.EUI64] = set()
        self._routes_unsupported: set[t.EUI64] = set()

        self.neighbors: dict[t.EUI64, list[zdo_t.Neighbor]] = collections.defaultdict(
            list
        )
        self.routes: dict[t.EUI64, list[zdo_t.Route]] = collections.defaultdict(list)

    def start_periodic_scans(self, period: float) -> None:
        self.stop_periodic_scans()
        self._scan_loop_task = asyncio.create_task(self._scan_loop(period))

    def stop_periodic_scans(self) -> None:
        if self._scan_loop_task is not None:
            self._scan_loop_task.cancel()

    async def _scan_loop(self, period: float) -> None:
        """Delay scan by creating a task."""

        while True:
            await asyncio.sleep(period)

            # Don't run a scheduled scan if a scan is already running
            if self._scan_task is not None and not self._scan_task.done():
                continue

            LOGGER.debug("Starting scheduled neighbor scan")

            try:
                await self.scan()
            except asyncio.CancelledError:
                # We explicitly catch a cancellation here to ensure the scan loop will
                # not be interrupted if a manual scan is initiated
                LOGGER.debug("Topology scan cancelled")
            except (Exception, asyncio.CancelledError):  # noqa: BLE001
                LOGGER.debug("Topology scan failed", exc_info=True)

    async def scan(
        self, devices: typing.Iterable[zigpy.device.Device] | None = None
    ) -> None:
        """Preempt Topology scan and reschedule."""

        if self._scan_task and not self._scan_task.done():
            LOGGER.debug("Cancelling old scanning task")
            self._scan_task.cancel()

        self._scan_task = asyncio.create_task(self._scan(devices))
        await self._scan_task

    async def _scan_table(
        self, scan_request: typing.Callable, entries_attr: str
    ) -> list[typing.Any]:
        """Scan a device table by sending ZDO requests."""

        index = 0
        table = []

        while True:
            status, rsp = await scan_request(index, retries=2, retry_delay=1.0)

            if status != zdo_t.Status.SUCCESS:
                raise ScanNotSupported

            entries = getattr(rsp, entries_attr)

            table.extend(entries)
            index += len(entries)

            # We intentionally sleep after every request, even the last one, to simplify
            # delay logic when scanning many devices in quick succession
            await asyncio.sleep(random.uniform(*REQUEST_DELAY))

            if index >= rsp.Entries or not entries:
                break

        return table

    async def _scan_neighbors(
        self, device: zigpy.device.Device
    ) -> list[zdo_t.Neighbor]:
        if device.ieee in self._neighbors_unsupported:
            return []

        LOGGER.debug("Scanning neighbors of %s", device)

        try:
            table = await self._scan_table(device.zdo.Mgmt_Lqi_req, "NeighborTableList")
        except ScanNotSupported:
            table = []
            self._neighbors_unsupported.add(device.ieee)

        return [n for n in table if n.ieee not in INVALID_NEIGHBOR_IEEES]

    async def _scan_routes(self, device: zigpy.device.Device) -> list[zdo_t.Route]:
        if device.ieee in self._routes_unsupported:
            return []

        LOGGER.debug("Scanning routing table of %s", device)

        try:
            table = await self._scan_table(device.zdo.Mgmt_Rtg_req, "RoutingTableList")
        except ScanNotSupported:
            table = []
            self._routes_unsupported.add(device.ieee)

        return table

    async def _scan(
        self, devices: typing.Iterable[zigpy.device.Device] | None = None
    ) -> None:
        """Scan topology."""

        if devices is None:
            # We iterate over a copy of the devices as opposed to the live dictionary
            devices = list(self._app.devices.values())

        for index, device in enumerate(devices):
            LOGGER.debug(
                "Scanning topology (%d/%d) of %s", index + 1, len(devices), device
            )

            # Ignore devices that aren't routers
            if device.node_desc is None or not (
                device.node_desc.is_router or device.node_desc.is_coordinator
            ):
                continue

            # Ignore devices that do not support scanning tables
            if (
                device.ieee in self._neighbors_unsupported
                and device.ieee in self._routes_unsupported
            ):
                continue

            # Some coordinators have issues when performing loopback scans
            if (
                self._app.config[zigpy.config.CONF_TOPO_SKIP_COORDINATOR]
                and device is self._app._device
            ):
                continue

            try:
                self.neighbors[device.ieee] = await self._scan_neighbors(device)
            except Exception as e:  # noqa: BLE001
                LOGGER.debug("Failed to scan neighbors of %s", device, exc_info=e)
            else:
                LOGGER.debug(
                    "Scanned neighbors of %s: %s", device, self.neighbors[device.ieee]
                )

            self.listener_event(
                "neighbors_updated", device.ieee, self.neighbors[device.ieee]
            )

            try:
                # Filter out inactive routes
                routes = await self._scan_routes(device)
                self.routes[device.ieee] = [
                    route
                    for route in routes
                    if route.RouteStatus != zdo_t.RouteStatus.Inactive
                ]
            except Exception as e:  # noqa: BLE001
                LOGGER.debug("Failed to scan routes of %s", device, exc_info=e)
            else:
                LOGGER.debug(
                    "Scanned routes of %s: %s", device, self.routes[device.ieee]
                )

            self.listener_event("routes_updated", device.ieee, self.routes[device.ieee])

        LOGGER.debug("Finished scanning neighbors for all devices")
        await self._find_unknown_devices(neighbors=self.neighbors, routes=self.routes)

    async def _find_unknown_devices(
        self,
        *,
        neighbors: dict[t.EUI64, list[zdo_t.Neighbor]],
        routes: dict[t.EUI64, list[zdo_t.Route]],
    ) -> None:
        """Discover unknown devices discovered during topology scanning"""
        # Build a list of unknown devices from the topology scan
        unknown_nwks = set()

        for neighbor in itertools.chain.from_iterable(neighbors.values()):
            try:
                self._app.get_device(nwk=neighbor.nwk)
            except KeyError:
                unknown_nwks.add(neighbor.nwk)

        for route in itertools.chain.from_iterable(routes.values()):
            # Ignore inactive or pending routes
            if route.RouteStatus != zdo_t.RouteStatus.Active:
                continue

            for nwk in (route.DstNWK, route.NextHop):
                try:
                    self._app.get_device(nwk=nwk)
                except KeyError:
                    unknown_nwks.add(nwk)

        # Try to discover any unknown devices
        for nwk in unknown_nwks:
            LOGGER.debug("Found unknown device nwk=%s", nwk)
            await self._app._discover_unknown_device(nwk)
            await asyncio.sleep(random.uniform(*REQUEST_DELAY))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from zigpy import types as t
from zigpy.endpoint import Endpoint
import zigpy.profiles.zha as zha_profile
from zigpy.util import ListenableMixin, LocalLogMixin
import zigpy.zcl
from zigpy.zcl import foundation

if TYPE_CHECKING:
    from zigpy.application import ControllerApplication

LOGGER = logging.getLogger(__name__)


class Group(ListenableMixin, dict):
    def __init__(
        self,
        group_id: int,
        name: str | None = None,
        groups: Groups | None = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self._groups: Groups = groups
        self._group_id: t.Group = t.Group(group_id)
        self._name: str = name
        self._endpoint: GroupEndpoint = GroupEndpoint(self)
        self._send_sequence = 0

        if groups is not None:
            self.add_listener(groups)

    def get_sequence(self) -> int:
        self._send_sequence = (self._send_sequence + 1) % 256
        return self._send_sequence

    def add_member(self, ep: Endpoint, suppress_event: bool = False) -> Group:
        if not isinstance(ep, Endpoint):
            raise ValueError(f"{ep} is not {Endpoint.__class__.__name__} class")  # noqa: TRY004
        if ep.unique_id in self:
            return self[ep.unique_id]
        self[ep.unique_id] = ep
        ep.member_of[self.group_id] = self
        if not suppress_event:
            self.listener_event("member_added", self, ep)
        return self

    def remove_member(self, ep: Endpoint, suppress_event: bool = False) -> Group:
        self.pop(ep.unique_id, None)
        ep.member_of.pop(self.group_id, None)
        if not suppress_event:
            self.listener_event("member_removed", self, ep)
        return self

    async def request(self, profile, cluster, sequence, data, *args, **kwargs):
        """Send multicast request."""
        await self.application.send_packet(
            t.ZigbeePacket(
                src_ep=self.application.get_endpoint_id(
                    cluster, is_server_cluster=False
                ),
                dst=t.AddrModeAddress(
                    addr_mode=t.AddrMode.Group, address=self.group_id
                ),
                tsn=sequence,
                profile_id=profile,
                cluster_id=cluster,
                data=t.SerializableBytes(data),
                radius=0,
                non_member_radius=3,
            )
        )

        return foundation.GENERAL_COMMANDS[
            foundation.GeneralCommand.Default_Response
        ].schema(
            status=foundation.Status.SUCCESS,
            command_id=data[2],
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} group_id={self.group_id} name='{self.name}' members={super().__repr__()}>"

    @property
    def application(self) -> ControllerApplication:
        """Expose application to FakeEndpoint/GroupCluster."""
        return self.groups.application

    @property
    def groups(self) -> Groups:
        return self._groups

    @property
    def group_id(self) -> t.Group:
        return self._group_id

    @property
    def members(self) -> Group:
        return self

    @property
    def name(self) -> str:
        if self._name is None:
            return f"No name group {self.group_id}"
        return self._name

    @property
    def endpoint(self) -> GroupEndpoint:
        return self._endpoint


class Groups(ListenableMixin, dict):
    def __init__(self, app: ControllerApplication, *args: Any, **kwargs: Any):
        self._application: ControllerApplication = app
        self._listeners: dict = {}
        super().__init__(*args, **kwargs)

    def add_group(
        self, group_id: int, name: str | None = None, suppress_event: bool = False
    ) -> Group:
        if group_id in self:
            return self[group_id]
        LOGGER.debug("Adding group: %s, %s", group_id, name)
        group = Group(group_id, name, self)
        self[group_id] = group
        if not suppress_event:
            self.listener_event("group_added", group)
        return group

    def member_added(self, group: Group, ep: Endpoint) -> None:
        self.listener_event("group_member_added", group, ep)

    def member_removed(self, group: Group, ep: Endpoint) -> None:
        self.listener_event("group_member_removed", group, ep)

    def pop(self, item, *args: Any) -> Group | None:
        if isinstance(item, Group):
            group = super().pop(item.group_id, *args)
            if isinstance(group, Group):
                for member in (*group.values(),):
                    group.remove_member(member)
                self.listener_event("group_removed", group)
            return group
        group = super().pop(item, *args)
        if isinstance(group, Group):
            for member in (*group.values(),):
                group.remove_member(member)
        self.listener_event("group_removed", group)
        return group

    remove_group = pop

    def update_group_membership(self, ep: Endpoint, groups: set[int]) -> None:
        """Sync up device group membership."""
        old_groups = {
            group.group_id for group in self.values() if ep.unique_id in group.members
        }

        for grp_id in old_groups - groups:
            self[grp_id].remove_member(ep)

        for grp_id in groups - old_groups:
            group = self.add_group(grp_id)
            group.add_member(ep)

    @property
    def application(self) -> ControllerApplication:
        """Return application controller."""
        return self._application


class GroupCluster(zigpy.zcl.Cluster):
    """Virtual cluster for group requests."""

    @classmethod
    def from_id(  # type: ignore[override]
        cls, group_endpoint: GroupEndpoint, cluster_id: int, is_server=True
    ) -> zigpy.zcl.Cluster:
        """Instantiate from ZCL cluster by cluster id."""
        if is_server is not True:
            raise ValueError("Only server clusters are supported for group requests")
        if cluster_id in cls._registry:
            return cls._registry[cluster_id](group_endpoint, is_server=True)
        group_endpoint.debug(
            "0x%04x cluster id is not supported for group requests", cluster_id
        )
        raise KeyError(f"Unsupported 0x{cluster_id:04x} cluster id for groups")

    @classmethod
    def from_attr(
        cls, group_endpoint: GroupEndpoint, ep_name: str
    ) -> zigpy.zcl.Cluster:
        """Instantiate by Cluster name."""

        for cluster in cls._registry.values():
            if cluster.ep_attribute == ep_name:
                return cluster(group_endpoint, is_server=True)
        raise AttributeError(f"Unsupported {ep_name} group cluster")


class GroupEndpoint(LocalLogMixin):
    """Group request handlers.

    wrapper for virtual clusters.
    """

    def __init__(self, group: Group):
        """Instantiate GroupRequest."""
        self._group: Group = group
        self._clusters: dict = {}
        self._cluster_by_attr: dict = {}

    @property
    def endpoint_id(self) -> None:
        return None

    @property
    def clusters(self) -> dict:
        """Group clusters.

        most of the times, group requests are addressed from client -> server clusters.
        """
        return self._clusters

    @property
    def device(self) -> Group:
        """Group is our fake zigpy device"""
        return self._group

    def request(self, cluster, sequence, data, *args, **kwargs):
        """Send multicast request."""
        return self.device.request(zha_profile.PROFILE_ID, cluster, sequence, data)

    def reply(self, cluster, sequence, data, *args, **kwargs):
        """Send multicast reply.

        do we really need this one :shrug:
        """
        return self.request(cluster, sequence, data, *args, **kwargs)

    def log(self, lvl: int, msg: str, *args: Any, **kwargs: Any) -> None:
        msg = "[0x%04x] " + msg
        args = (self._group.group_id, *args)
        LOGGER.log(lvl, msg, *args, **kwargs)

    def __getitem__(self, item: int):
        """Return or instantiate a group cluster."""
        try:
            return self.clusters[item]
        except KeyError:
            self.debug("trying to create new group %s cluster id", item)

        cluster = GroupCluster.from_id(self, item)
        self.clusters[item] = cluster
        return cluster

    def __getattr__(self, name: str):
        """Return or instantiate a group cluster by cluster name."""
        try:
            return self._cluster_by_attr[name]
        except KeyError:
            self.debug("trying to create a new group '%s' cluster", name)

        cluster = GroupCluster.from_attr(self, name)
        self._cluster_by_attr[name] = cluster
        return cluster



"""Compatibility shim: number device classes moved into ZHA."""

from __future__ import annotations

import importlib
import typing
import warnings


def __getattr__(name: str) -> typing.Any:
    warnings.warn(
        f"`{__name__}.{name}` is deprecated, import it from"
        " `zha.application.platforms.number.device_class` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(
        importlib.import_module("zha.application.platforms.number.device_class"), name
    )



from __future__ import annotations

from collections.abc import Coroutine
import functools
import logging

from zigpy.const import APS_REPLY_TIMEOUT
import zigpy.profiles
import zigpy.types as t
import zigpy.util

from . import types

LOGGER = logging.getLogger(__name__)

ZDO_ENDPOINT = 0


class ZDO(zigpy.util.CatchingTaskMixin, zigpy.util.ListenableMixin):
    """The ZDO endpoint of a device"""

    class LeaveOptions(t.bitmap8):
        """ZDO Mgmt_Leave_req Options."""

        NONE = 0
        RemoveChildren = 1 << 6
        Rejoin = 1 << 7

    def __init__(self, device):
        self._device = device
        self._listeners = {}

    def _serialize(self, command, *args, **kwargs):
        keys, schema = types.CLUSTERS[command]
        # TODO: expose this in a future PR
        assert not kwargs
        return t.serialize(args, schema)

    def deserialize(self, cluster_id, data):
        if cluster_id not in types.CLUSTERS:
            raise ValueError(f"Invalid ZDO cluster ID: 0x{cluster_id:04X}")

        _, param_types = types.CLUSTERS[cluster_id]
        hdr, data = types.ZDOHeader.deserialize(cluster_id, data)
        args, data = t.deserialize(data, param_types)

        if data:
            # TODO: Seems sane to check, but what should we do?
            self.warning("Data remains after deserializing ZDO frame: %r", data)

        return hdr, args

    async def request(
        self,
        command,
        *args,
        timeout=APS_REPLY_TIMEOUT,
        expect_reply: bool = True,
        use_ieee: bool = False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_delay: float | None = None,
        **kwargs,
    ):
        data = self._serialize(command, *args, **kwargs)
        tsn = self.device.get_sequence()
        return await self._device.request(
            profile=0x0000,
            cluster=command,
            src_ep=ZDO_ENDPOINT,
            dst_ep=ZDO_ENDPOINT,
            sequence=tsn,
            data=t.uint8_t(tsn).serialize() + data,
            timeout=timeout,
            expect_reply=expect_reply,
            use_ieee=use_ieee,
            ask_for_ack=ask_for_ack,
            priority=priority,
            retries=retries,
            retry_delay=retry_delay,
        )

    async def reply(
        self,
        command,
        *args,
        tsn: int | t.uint8_t | None = None,
        timeout=APS_REPLY_TIMEOUT,
        expect_reply: bool = False,
        use_ieee: bool = False,
        ask_for_ack: bool | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_delay: float | None = None,
        **kwargs,
    ):
        data = self._serialize(command, *args, **kwargs)
        if tsn is None:
            tsn = self.device.get_sequence()
        return await self._device.reply(
            profile=0x0000,
            cluster=command,
            src_ep=ZDO_ENDPOINT,
            dst_ep=ZDO_ENDPOINT,
            sequence=tsn,
            data=t.uint8_t(tsn).serialize() + data,
            timeout=timeout,
            expect_reply=expect_reply,
            use_ieee=use_ieee,
            ask_for_ack=ask_for_ack,
            priority=priority,
            retries=retries,
            retry_delay=retry_delay,
        )

    def handle_message(
        self,
        profile: int,
        cluster: int,
        hdr: types.ZDOHeader,
        args: list,
    ) -> None:
        self.debug("ZDO request %s: %s", hdr.command_id, args)

        handler = getattr(self, f"handle_{hdr.command_id.name.lower()}", None)
        if handler is not None:
            handler(hdr, *args)
        else:
            self.debug("No handler for ZDO request:%s(%s)", hdr.command_id, args)

        self.listener_event(
            f"zdo_{hdr.command_id.name.lower()}",
            self._device,
            None,  # was `dst_addressing` but this was never set
            hdr,
            args,
        )

    def handle_nwk_addr_req(
        self,
        hdr: types.ZDOHeader,
        ieee: t.EUI64,
        request_type: int,
        start_index: int | None = None,
    ):
        """Handle ZDO NWK Address request."""

        app = self._device.application
        if ieee == app.state.node_info.ieee:
            self.create_catching_task(
                self.NWK_addr_rsp(
                    0,
                    app.state.node_info.ieee,
                    app.state.node_info.nwk,
                    0,
                    0,
                    [],
                    tsn=hdr.tsn,
                    priority=t.PacketPriority.LOW,
                )
            )

    def handle_ieee_addr_req(
        self,
        hdr: types.ZDOHeader,
        nwk: t.NWK,
        request_type: int,
        start_index: int | None = None,
    ):
        """Handle ZDO IEEE Address request."""

        app = self._device.application
        if nwk in (
            t.BroadcastAddress.ALL_DEVICES,
            t.BroadcastAddress.RX_ON_WHEN_IDLE,
            t.BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR,
            app.state.node_info.nwk,
        ):
            self.create_catching_task(
                self.IEEE_addr_rsp(
                    0,
                    app.state.node_info.ieee,
                    app.state.node_info.nwk,
                    0,
                    0,
                    [],
                    tsn=hdr.tsn,
                    priority=t.PacketPriority.LOW,
                )
            )

    def handle_device_annce(
        self,
        hdr: types.ZDOHeader,
        nwk: t.NWK,
        ieee: t.EUI64,
        capability: int,
    ):
        """Handle ZDO device announcement request."""
        self.listener_event("device_announce", self._device)

    def handle_mgmt_permit_joining_req(
        self,
        hdr: types.ZDOHeader,
        permit_duration: int,
        tc_significance: int,
    ):
        """Handle ZDO permit joining request."""

        self.listener_event("permit_duration", permit_duration)

    def handle_match_desc_req(
        self,
        hdr: types.ZDOHeader,
        addr: t.NWK,
        profile: int,
        in_clusters: list,
        out_cluster: list,
    ):
        """Handle ZDO Match_desc_req request."""

        local_addr = self._device.application.state.node_info.nwk
        if profile != zigpy.profiles.zha.PROFILE_ID:
            self.create_catching_task(
                self.Match_Desc_rsp(
                    0,
                    local_addr,
                    [],
                    tsn=hdr.tsn,
                    priority=t.PacketPriority.CRITICAL,
                )
            )
            return

        self.create_catching_task(
            self.Match_Desc_rsp(
                0,
                local_addr,
                [t.uint8_t(1)],
                tsn=hdr.tsn,
                priority=t.PacketPriority.CRITICAL,
            )
        )

    async def bind(self, cluster, **kwargs):
        return await self.Bind_req(
            self._device.ieee,
            cluster.endpoint.endpoint_id,
            cluster.cluster_id,
            self.device.application.get_dst_address(cluster),
            **kwargs,
        )

    async def unbind(self, cluster):
        return await self.Unbind_req(
            self._device.ieee,
            cluster.endpoint.endpoint_id,
            cluster.cluster_id,
            self.device.application.get_dst_address(cluster),
        )

    def leave(self, remove_children: bool = True, rejoin: bool = False) -> Coroutine:
        opts = self.LeaveOptions.NONE
        if remove_children:
            opts |= self.LeaveOptions.RemoveChildren
        if rejoin:
            opts |= self.LeaveOptions.Rejoin

        return self.Mgmt_Leave_req(self._device.ieee, opts)

    def permit(self, duration=60, tc_significance=0):
        return self.Mgmt_Permit_Joining_req(duration, tc_significance)

    def log(self, lvl, msg, *args, **kwargs):
        msg = "[0x%04x:zdo] " + msg
        args = (self._device.nwk, *args)
        return LOGGER.log(lvl, msg, *args, **kwargs)

    @property
    def device(self):
        return self._device

    def __getattr__(self, name):
        try:
            command = types.ZDOCmd[name]
        except KeyError as exc:
            raise AttributeError(f"No such '{name}' ZDO command") from exc

        if command & 0x8000:
            return functools.partial(self.reply, command)
        return functools.partial(self.request, command)


def broadcast(
    app,
    command,
    grpid,
    radius,
    *args,
    broadcast_address=t.BroadcastAddress.RX_ON_WHEN_IDLE,
    **kwargs,
):
    params, param_types = types.CLUSTERS[command]

    named_args = dict(zip(params, args, strict=False))
    named_args.update(kwargs)
    assert set(named_args.keys()) == set(params)

    sequence = app.get_sequence()
    data = bytes([sequence]) + t.serialize(named_args.values(), param_types)

    return zigpy.device.broadcast(
        app,
        0,
        command,
        0,
        0,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=broadcast_address,
    )



from __future__ import annotations

from zigpy.types import basic


class DeviceID(basic.uint32_t, repr="hex"):
    pass


class FrameType(basic.enum2):
    DataFrame = 0x00
    MaintenanceFrame = 0x01


class ApplicationID(basic.enum3):
    SrcID = 0b000
    IEEE = 0b010
    LPED = 0b001


# Table 13
class SecurityLevel(basic.enum2):
    NoSecurity = 0b00
    ShortFrameCounterAndMIC = 0b01
    FullFrameCounterAndMIC = 0b10
    Encrypted = 0b11


# Table 14
class SecurityKeyType(basic.enum3):
    NoKey = 0b000
    NWKKey = 0b001
    GPDGroupKey = 0b010
    NWKKeyDerivedGPD = 0b011
    IndividualKey = 0b100
    DerivedIndividual = 0b111


# ZGP spec Figure 22
class ProxyCommissioningModeExitMode(basic.enum3):
    NotDefined = 0b000
    OnExpire = 0b001
    OnFirstPairing = 0b010
    OnExplicitExit = 0b100
    OnExpireOrFirstPairing = 0b011
    OnExpireOrExplicitExit = 0b101


# Table 29
class CommunicationMode(basic.enum2):
    Unicast = 0b00
    GroupcastForwardToDGroup = 0b01
    GroupcastForwardToCommGroup = 0b10
    UnicastLightweight = 0b11


class CommunicationDirection(basic.enum1):
    GPDtoGPP = 0
    GPPtoGPD = 1



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import zigpy.backups


class ZigbeeException(Exception):
    """Base exception class"""


class ParsingError(ZigbeeException):
    """Failed to parse a frame"""


class ControllerException(ZigbeeException):
    """Application controller failed in some way."""


class CannotWriteNetworkSettings(ZigbeeException):
    """The provided network settings cannot be written due to a radio limitation."""


class DestructiveWriteNetworkSettings(ZigbeeException):
    """The provided network settings will be written but in a destructive manner."""


class APIException(ZigbeeException):
    """Radio API failed in some way."""


class DeliveryError(ZigbeeException):
    """Message delivery failed in some way"""

    def __init__(self, message: str, status: int | None = None):
        super().__init__(message)
        self.status = status


class SendError(DeliveryError):
    """Message could not be enqueued."""


class InvalidResponse(ZigbeeException):
    """A ZDO or ZCL response has an unsuccessful status code"""


class RadioException(Exception):
    """Base exception class for radio exceptions"""


class TransientConnectionError(RadioException):
    """Connection to the radio failed but will likely succeed in the near future"""


class NetworkNotFormed(RadioException):
    """A network cannot be started because the radio has no stored network info"""


class FormationFailure(RadioException):
    """Network settings could not be written to the radio"""


class NetworkSettingsInconsistent(ZigbeeException):
    """Loaded network settings are different from what is in the database"""

    def __init__(
        self,
        message: str,
        new_state: zigpy.backups.NetworkBackup,
        old_state: zigpy.backups.NetworkBackup,
    ) -> None:
        super().__init__(message)
        self.new_state = new_state
        self.old_state = old_state


class CorruptDatabase(ZigbeeException):
    """The SQLite database is corrupt or otherwise inconsistent"""


class QuirksException(Exception):
    """Base exception class"""


class MultipleQuirksMatchException(QuirksException):
    """Thrown when multiple v2 quirks match a device"""
